"""
This script is a key part of the following publications:
    - Herig Coimbra, Pedro Henrique and Loubet, Benjamin and Laurent, Olivier and Mauder, Matthias and Heinesch, Bernard and 
    Bitton, Jonathan and Delpierre, Nicolas and Depuydt, Jérémie and Buysse, Pauline, Improvement of Co2 Flux Quality Through 
    Wavelet-Based Eddy Covariance: A New Method for Partitioning Respiration and Photosynthesis. 
    Available at SSRN: https://ssrn.com/abstract=4642939 or http://dx.doi.org/10.2139/ssrn.4642939
"""

##########################################
###     IMPORTS                           
##########################################

# standard modules
import copy
import os
import re
import warnings
import logging
import time
import datetime
from functools import reduce

# 3rd party modules
import yaml
import numpy as np
from itertools import chain
from scipy.optimize import curve_fit
from sklearn.linear_model import LinearRegression, RANSACRegressor, HuberRegressor, QuantileRegressor
import zipfile
from io import StringIO

# project modules
from .addons import *

##########################################
###     PROJECT CHOICES                           
##########################################

SITES_TO_STUDY = ['SAC']

month2season = lambda month: {1:1, 2:1, 3:2, 4:2, 5:2, 6:3, 7:3, 8:3, 9:4, 10:4, 11:4, 12:1}[month]


##########################################
###     GENERIC FUNCTIONS                           
##########################################


class structuredData:
    def __init__(self, **kwargs):
        for k, v in kwargs.items(): self.__dict__[k]=v
        pass


def matrixtotimetable(time, mat, c0name="TIMESTAMP", **kwargs):
    assert len(time) in mat.shape, f"Time ({time.shape}) and matrix ({mat.shape}) do not match."
    mat = np.array(mat)

    if len(time) != mat.shape[0] and len(time) == mat.shape[1]:
        mat = mat.T

    __temp__ = pd.DataFrame(mat, **kwargs)
    __temp__.insert(0, c0name, time)

    return __temp__


def yaml_to_dict(path):
    with open(path, 'r+') as file:
        file = yaml.safe_load(file)
    return file


def list_time_in_period(tmin, tmax, fastfreq, slowfreq, include='both'):
    if include=="left":
        return [(pd.date_range(max(p, pd.to_datetime(tmin)), min(p + pd.Timedelta(slowfreq), pd.to_datetime(tmax)),
          freq=fastfreq)[:-1]) for p in pd.date_range(tmin, tmax, freq=slowfreq).floor(slowfreq)]
    elif include == "right":
        return [(pd.date_range(max(p, pd.to_datetime(tmin)), min(p + pd.Timedelta(slowfreq), pd.to_datetime(tmax)),
          freq=fastfreq)[1:]) for p in pd.date_range(tmin, tmax, freq=slowfreq).floor(slowfreq)]
    elif include == "both":
        return [(pd.date_range(max(p, pd.to_datetime(tmin)), min(p + pd.Timedelta(slowfreq), pd.to_datetime(tmax)),
          freq=fastfreq)) for p in pd.date_range(tmin, tmax, freq=slowfreq).floor(slowfreq)]
    return


def checkifinprogress(path, LIMIT_TIME_OUT=30*60):
    if os.path.exists(path) and (time.time()-os.path.getmtime(path)) < LIMIT_TIME_OUT:
        logging.debug(f'Fresh file found ({time.time()-os.path.getmtime(path)} s old, {os.path.getmtime(path)}), skipping.')
        return 1
    else:
        if os.path.exists(path): logging.debug(f'Old file found ({time.time()-os.path.getmtime(path)} s old, {time.time()*10**-3}, {os.path.getmtime(path)*10**-3}), new in progress file created.')
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "a+"):
            pass
        return 0


def mkdirs(filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)


def nearest(items, pivot, direction=0):
    if direction == 0:
        nearest = min(items, key=lambda x: abs(x - pivot))
        difference = abs(nearest - pivot)
        
    elif direction == -1:
        nearest = min(items, key=lambda x: abs(x - pivot) if x<pivot else pd.Timedelta(999, "d"))
        difference = (nearest - pivot)
        
    elif direction == 1:
        nearest = min(items, key=lambda x: abs(x - pivot) if x>pivot else pd.Timedelta(999, "d"))
        difference = (nearest - pivot)
    return nearest, difference


def update_nested_dict(d, u):
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = update_nested_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def update_nested_dicts(*ds, fstr=None):
    r = {}
    for d in ds:
        if isinstance(d, str) and fstr:
            try:
                d = fstr(d)
            except Exception as e:
                continue
        r = update_nested_dict(r, d)
    return r


def concat_into_single_file(path, pattern, output_path=None, **kwargs):
    print('\nCONSOLIDATING DATASET\n')
    if output_path is None: output_path = os.path.join(path, 'concat_into_single_file') 
    
    files_to_concat = []
    for name in os.listdir(path):
        if re.findall(pattern, name):
            files_to_concat += [os.path.join(path, name)]
    
    files_to_concat = [pd.read_csv(f, **kwargs) for f in files_to_concat]
    data = reduce(lambda left, right: pd.concat([left, right]), files_to_concat)
    
    mkdirs(output_path)
    data.to_csv(output_path, index=False)
    print(os.path.basename(output_path), ': Saved.', ' '*15, end='\n', sep='')
    
    return


def __input_to_series__(data, request):
    if data is None: return request

    columns = data.columns
    if request is None: return 0 
    elif isinstance(request, str) and request in columns: 
        request = data[request]
    elif isinstance(request, (list, tuple)) and all(isinstance(i, str) and (i in columns) for i in request): 
        request = np.sum(data[request], axis=1)
    return request


def force_array_dimension(shape, out):
    """
    Return array with same shape as base one.
    """
    out_ = np.zeros(shape) * np.nan

    shape_dif = (np.array(shape) - np.array(out.shape)) / 2
    signal = shape_dif / abs(shape_dif)

    bas_cut = [None] * 4
    out_cut = [None] * 4

    for i, s_ in enumerate(signal):

        dif = [int(np.ceil(abs(shape_dif[i]))), -
               int(np.floor(abs(shape_dif[i])))]
        dif = [el if el != 0 else None for el in dif]

        if i == 0:
            if s_ == 1:
                bas_cut[:2] = dif

            elif s_ == -1:
                out_cut[:2] = dif

        elif i == 1:
            if s_ == 1:
                bas_cut[2:] = dif

            elif s_ == -1:
                out_cut[2:] = dif

    out_[bas_cut[0]:bas_cut[1],
         bas_cut[2]:bas_cut[3]] = \
        sum_nan_arrays(out_[bas_cut[0]:bas_cut[1],
                               bas_cut[2]:bas_cut[3]],
                          out[out_cut[0]:out_cut[1],
                              out_cut[2]:out_cut[3]])

    return out_


##########################################
###     STATISTICS (PRINT)
##########################################


def custom_round(x, base=5):
    return base * np.round(x/base)


def nanminmax(x):
    return [np.nanmin(x), np.nanmax(x)]


def get_r2(X, y):
    if len(X)==0:
        return 0
    X = np.array(X).ravel()
    y = np.array(y).ravel()
    finite = np.isfinite(X*y)
    X = X[finite].reshape(-1, 1)
    y = y[finite].reshape(-1, 1)
    regression = LinearRegression(fit_intercept=True)
    regression.fit(X, y)
    r2 = regression.score(X, y)
    return r2


def __fit_noise__(y_axis, x_axis, x_max=5, a=1):
    logger = logging.getLogger('fitnoise')
    """
    freqs = [1/j2sj(j, 1/dt) for j in np.arange(1,len(spec)+1)]
    a: noise color (white = 1)
    """
    r = structuredData()
    r.curve_0 = lambda f, b: np.log((f**a)*b)
    specna = np.where(np.isnan(y_axis[:x_max]) | (y_axis[:x_max] <= 0), False, True)
    try:
        x_axis = np.array(x_axis)
        y_axis  = np.array(y_axis)
        r.params_0, _ = curve_fit(r.curve_0, x_axis[:x_max][specna], np.log(y_axis[:x_max][specna]), bounds=(0, np.inf))
    except Exception as e:
        logger.error(f"UserWarning: {str(e)}")
        r.params_0 = [0]
    r.fit = np.array([(f**a)*r.params_0[0] for f in x_axis])
    return r


def sum_nan_arrays(a, b):
    ma = np.isnan(a)
    mb = np.isnan(b)
    return np.where(ma & mb, np.nan, np.where(ma, 0, a) + np.where(mb, 0, b))


def summarisestats(X, y, method='Linear', fit_intercept=True, **kw):
    statisticsToReturn = structuredData()
    X = np.array(X)
    y = np.array(y)
    NaN = np.isnan(X) + np.isnan(y)
    X = X[NaN==0]
    y = y[NaN==0]

    regression = {'Linear': LinearRegression,
                  'RANSAC': RANSACRegressor,
                  'Huber':  HuberRegressor}[method](fit_intercept=fit_intercept, **kw)
    regression.fit(X.reshape(-1, 1), y.reshape(-1, 1))
    statisticsToReturn.m = regression.coef_[0]#[0]
    b = regression.intercept_
    statisticsToReturn.b = b
    b_ = np.round(b, 2) if b else 0
    b_ = "+" + str(b_) if b_ >= 0 else str(b_)
    if method == 'Huber':
        statisticsToReturn.outliers = regression.outliers_
        print(len(regression.outliers_), sum(regression.outliers_))
        X = X[regression.outliers_==False]
        y = y[regression.outliers_==False]
        
    statisticsToReturn.me = np.nanmean((X-y)).round(2)
    statisticsToReturn.mae = np.nanmean(abs(X-y)).round(2)
    statisticsToReturn.r2 = regression.score(X.reshape(-1, 1), y.reshape(-1, 1))
    return statisticsToReturn

def summarisestatslabel(meta, xn, yn):    
    stat_label = f"R²: {np.round(meta.r2, 2)}"
    stat_label = stat_label + f"\nME: {np.round(meta.me, 2)}"
    stat_label = stat_label + f"\nMAE: {np.round(meta.mae, 2)}"
    stat_label = stat_label + f"\n{yn}={np.round(meta.m, 2)} {xn}"
    return stat_label
    
def summarisestatstext(meta, xn='x', yn='y'):    
    stat_label = f"R²= {np.round(meta.r2, 2)}"
    stat_label = stat_label + f", ME= {np.round(meta.me, 2)} µmol m-2 s-1"
    stat_label = stat_label + f", MAE= {np.round(meta.mae, 2)} µmol m-2 s-1"
    stat_label = stat_label + f", {yn}={np.round(meta.m, 2)}"+r"$\times$"+"{xn} linear fit" #×
    return stat_label


##########################################
###     PLOT                           
##########################################


def abline(intercept=0, slope=1, origin=(0, 0), **kwargs):
    kwargs.pop('data', None)
    axes = plt.gca()
    x_vals = np.array(axes.get_xlim())
    y_vals = intercept + slope * x_vals
    plt.plot(x_vals, y_vals, **kwargs)


def abline2(origin=(0, 0), slope=1, length=1, scale='', **kwargs):
    kwargs.pop('data', None)
    if scale=='log': origin = np.log10(origin)
    x_vals = np.array([origin[0], origin[0]+length])#np.linspace(origin[0], origin[0]+length)
    y_vals = origin[1] - slope * x_vals[0] + slope * x_vals
    if scale=='log': 
        x_vals = 10**x_vals
        y_vals = 10**y_vals
    plt.plot(x_vals, y_vals, **kwargs)


def add_subplot_axes(ax,rect):
    fig = plt.gcf()
    box = ax.get_position()
    width = box.width
    height = box.height
    inax_position  = ax.transAxes.transform(rect[0:2])
    transFigure = fig.transFigure.inverted()
    infig_position = transFigure.transform(inax_position)    
    x = infig_position[0]
    y = infig_position[1]
    width *= rect[2]
    height *= rect[3]  # <= Typo was here
    #subax = fig.add_axes([x,y,width,height],facecolor=facecolor)  # matplotlib 2.0+
    subax = fig.add_axes([x,y,width,height])
    x_labelsize = subax.get_xticklabels()[0].get_size()
    y_labelsize = subax.get_yticklabels()[0].get_size()
    x_labelsize *= rect[2]**0.5
    y_labelsize *= rect[3]**0.5
    subax.xaxis.set_tick_params(labelsize=x_labelsize)
    subax.yaxis.set_tick_params(labelsize=y_labelsize)
    return subax

##########################################
###     GET DATASETS                           
##########################################

def get_all_sites(fc, *args, **kwargs):
    datasetToReturn = structuredData()
    datasetToReturn.data = {}
    for sitename in SITES_TO_STUDY:
        data = fc(*args, sitename=sitename, **kwargs)
        if data is None: data = pd.DataFrame({"TIMESTAMP": []})
        datasetToReturn.data[sitename] = data
    datasetToReturn.alldata = []
    for k, v in datasetToReturn.data.items():
        v.insert(0, 'CO_SITE', k)
        datasetToReturn.alldata += [copy.deepcopy(v).reset_index(drop=True)]
    datasetToReturn.alldata = pd.concat(
        datasetToReturn.alldata).reset_index(drop=True)
    return datasetToReturn.alldata

def get_cospectra(sitename=None, **kwargs):
    if sitename is None:
        return get_all_sites(get_cospectra, **kwargs)
    mergeorconcat = kwargs.get('mergeorconcat', 'merge')
    folder        = kwargs.get('folder', 'data')
    duplicates    = kwargs.get('duplicates', False)

    wv_path = os.path.join(folder, sitename, 'output', 'DWCS')
    data = []
    for name in os.listdir(wv_path):
        if any([name.endswith(ext) for ext in ['csv', 'xlsx', 'txt', 'parquet', 'json']]):
            if re.findall('_full_cospectra', name) and re.findall('.30mn', name):
                data.append(pd.read_file(os.path.join(wv_path, name)))
    
    if len(data) == 0:
        return None
    elif len(data) == 1:
        data = data[0]
    elif mergeorconcat == 'concat':
        data = pd.concat(data)
    else:
        if duplicates:
            data = reduce(lambda left, right: pd.merge(left, right, on=['TIMESTAMP'], how="outer", suffixes=('', '_DUP')),
                            data)
        else:
            print(data)
            data = reduce(lambda left, right: pd.merge(left, right[['TIMESTAMP'] + list(set(right) - set(left))], 
                                                       on=['TIMESTAMP'], how="outer", suffixes=('', '_DUP')), data)
    
    for tc in data.columnstartswith('TIMESTAMP'):
        data[tc] = pd.to_datetime(
            data[tc])
    return data
    
def get_metadata(sitename=None, folder='data'):
    if sitename is None:
        return get_all_sites(get_metadata)
    
    mt_path = os.path.join(folder, sitename, f'{sitename}_metadata.yaml')
    if os.path.exists(mt_path):
        meta = yaml.safe_load(open(mt_path, 'r'))
        return pd.DataFrame(meta, index=[0])
    return None


def j2sj(e, samp_rate=10): return 1/(samp_rate*(2**-float(e)))
def sj2j(s, samp_rate=10): return np.log(samp_rate*s) / np.log(2)

def get_dic_flux_data(data): return {k: copy.deepcopy(data.query(
    f"CO_SITE == '{k}'").reset_index(drop=True)) for k in data.CO_SITE.unique()}


##########################################
###     UNIVERSAL READER                            
##########################################

DEFAULT_FILE_RAW = {
    'file_pattern': '([0-9]{8}-[0-9]{4})_raw_dataset_.*.txt', 
    'date_format': '%Y%m%d-%H%M', 
    'dt': 0.05, 
    'tname': "TIMESTAMP", 
    'id': None,
    'datefomatfrom': '%Y%m%d%H%M%S.%f', 
    'datefomatto': '%Y-%m-%dT%H:%M:%S.%f'
}

DEFAULT_READ_CSV = {
    'skiprows': 8,
    'sep': r"\s+",
    'na_values': ['NaN', 'nan', -9999],
}

DEFAULT_READ_GHG = {
    'skiprows': 7,
    'sep': r"\t",
    'engine': 'python'
}

DEFAULT_FMT_DATA = {
    'co': '4th',
}


class structuredDataFrame:
    def __init__(self, data=None, dt=None, **kwargs):
        if data is None:
            loopvar = kwargs.pop('lookup', [])
            loopvar = [l.to_list() if isinstance(l, list)==False else l for l in loopvar]
            for l in loopvar:
                result = universal_reader(lookup=l, **kwargs, verbosity=0)
                self.__dict__.update(result.__dict__)
        
        else:
            assert dt is not None, 'Missing measurement frequency (dt).'
            self.data = data
            self.dt = dt
            self.__dict__.update(**kwargs)
    
    def filter(self, items: dict):
        for k, v in items.items():
            if isinstance(v, tuple):
                self.data = self.data.loc[(self.data[k] > v[0])
                                          & (self.data[k] < v[1])].copy()
            else:
                self.data = self.data[self.data[k].isin(v)].copy()
        return self

    def rename(self, names: dict):
        self.data = self.data.rename(columns=names)
        return self

    def modify(self, items: dict):
        for k, v in items.items():
            self.data[k] = v
        return self

    def format(self, 
               cols={'t':'ts'}, 
               keepcols=['u', 'v', 'w', 'ts', 'co2', 'co2_dry', 'h2o', 'h2o_dry', 'ch4', 'n2o'],
               addkeep=[],
               colsfunc=str.lower, cut=False, **kwargs):
        
        if isinstance(self, pd.DataFrame):
            formated = self
        else:
            fmt_clas = structuredDataFrame(**self.__dict__)
            formated = fmt_clas.data

        if colsfunc is not None:
            formated.columns = map(colsfunc, formated.columns)
        #cols.update(kwargs)
        cols.update({v.lower(): k.lower() for k, v in kwargs.items() if isinstance(v, list)==False})
        cols = {v: k for k, v in {v: k for k, v in cols.items()}.items()}
        cols.update({'timestamp': 'TIMESTAMP'})
        #formated.TIMESTAMP = formated.TIMESTAMP.apply(np.datetime64)
        if cut:
            #formated = formated[[
            #    c for c in formated.columns if c in cols.keys()]]
            formated = formated.loc[:, np.isin(formated.columns, keepcols+addkeep+list(cols.keys()))]
        
        formated = formated.rename(columns=cols)

        if isinstance(self, pd.DataFrame):
            return formated
        else:
            fmt_clas.data = formated
            return fmt_clas
    
    def interpolate(self, cols=["co2", "w"], qcname="qc"):
        interpolated = structuredDataFrame(**self.__dict__)
        interpolated.data[qcname] = 0
        for c_ in list(cols):
            interpolated.data[qcname] = interpolated.data[qcname] + 0 * \
                np.array(interpolated.data[c_])
            interpolated.data.loc[np.isnan(interpolated.data[qcname]), qcname] = 1
            interpolated.data[qcname] = interpolated.data[qcname].astype(int)
            interpolated.data[c_] = interpolated.data[c_].interpolate(method='pad')
            
        return interpolated


def __universal_reader__(path, **kw_csv):
    if path.endswith('.gz'): kw_csv.update(**{'compression': 'gzip'})
    elif path.endswith('.csv'): kw_csv.pop('compression', None)
    if path.endswith('.ghg'):
        with zipfile.ZipFile(path, 'r') as zip_ref:
            datafile = [zip_ref.read(name) for name in zip_ref.namelist() if name.endswith(".data")][0]
        datafile = str(datafile, 'utf-8')
        path = StringIO(datafile)
        # DEFAULT_READ_GHG
        kw_csv.update(DEFAULT_READ_GHG)
    try:
        df_td = pd.read_csv(path, **kw_csv)
    except Exception as e:
        # (EOFError, pd.errors.ParserError, pd.errors.EmptyDataError):
        try:
            #if verbosity>1: warnings.warn(f'{e}, when opening {path}, using {kw_csv}. Re-trying using python as engine and ignoring bad lines.')
            df_td = pd.read_csv(path, on_bad_lines='warn', engine='python', **kw_csv)
        except Exception as ee:
            warnings.warn(f'{ee}, when opening {str(path)}, using {kw_csv}')
            return None
    return df_td

def universal_reader(path, lookup=[], fill=False, fmt={}, onlynumeric=True, verbosity=1, fkwargs={}, tipfile="readme.txt", **kwargs):
    """
    path: str or dict 
    e.g. (str): path = str(os.path.join(inputpath, 'eddypro_raw_datasets/level_6'))
    e.g. (dict): path = {k: str(os.path.join(v, 'eddypro_raw_datasets/level_6')) for k, v in inputpath.items()}
        where k will become the suffix in case of duplicates
    """
    if isinstance(path, dict):
        dataframes = {}
        for k, v in path.items():
            assert isinstance(v, str), f'Path ({v}) is not string.'
            path_ = universal_reader(v, lookup, fill, fmt, onlynumeric, verbosity, fkwargs, tipfile, **kwargs)
            dt = path_.dt
            dataframes[k] = path_.data
        #df_site = dataframes.pop(k, pd.DataFrame())
        dup_keys = {}
        for k in dataframes.keys():
            dup_keys[k] = [set(dataframes[k].columns).intersection(set(v_.columns)) - set(['TIMESTAMP']) for k_, v_ in dataframes.items() if k_!=k]
            dup_keys[k] = list(chain.from_iterable(dup_keys[k]))
            dataframes[k].rename(columns={c: c + k for c in dup_keys[k]}, inplace=True)
        
        try:
            df_site = reduce(lambda left, right: pd.merge(left, right, on='TIMESTAMP', how='outer', suffixes=('', '_DUP')), 
                            list(dataframes.values()))
        except Exception as e:
            logging.error(str(e))
            logging.debug('Concatenating instead of merging.')
            #df_site = reduce(lambda left, right: pd.concat([left, right], axis=1), dataframes)
            return structuredDataFrame(pd.DataFrame(), dt=dt)
        #for k, v in dataframes.items():
        #    df_site = df_site.merge(v, on='TIMESTAMP', how='outer', suffixes=('', k))
        return structuredDataFrame(df_site, dt=dt)
    
    df_site = pd.DataFrame()
    
    folders = [path + p + '/' for p in os.listdir(path) if os.path.isdir(path + p)]
    folders = folders if folders else [path]
    
    #print("Check readable files in", path if len(path)<40 else f'{path[:5]}...{path[-30:]}')#, fmt, fkwargs, kwargs)

    for path_ in folders:
        df_td = pd.DataFrame()

        # read tips file
        kw_ = update_nested_dicts({"FILE_RAW": DEFAULT_FILE_RAW, "READ_CSV": DEFAULT_READ_CSV, "FMT_DATA": DEFAULT_FMT_DATA}, 
                                  os.path.join(path, tipfile), os.path.join(path_, tipfile),
                                  {"FILE_RAW": fkwargs, "READ_CSV": kwargs, "FMT_DATA": fmt},
                                  fstr=lambda d: yaml_to_dict(d))
        
        kw = structuredData(**kw_['FILE_RAW'])
        kw_csv = kw_['READ_CSV']
        
        try:
            if ('header_file' in kw_csv.keys()) and (os.path.exists(kw_csv['header_file'])):
                kw_csv['header_file'] = "[" + open(kw_csv['header_file']).readlines()[0].replace("\n", "") + "]"
        except:
            None
        
        lookup_ = list(set([f.strftime(kw.date_format) for f in lookup]))
        files_list = {}

        for root, directories, files in os.walk(path_):
            for name in files:
                dateparts = re.findall(kw.file_pattern, name, flags=re.IGNORECASE)
                if len(dateparts) == 1:
                    files_list[dateparts[0]] = os.path.join(root, name)

        for td in set(lookup_) & files_list.keys() if lookup_ != [] else files_list.keys():
            path_to_tdfile = files_list[td]
            if os.path.exists(path_to_tdfile):
                df_td = __universal_reader__(path_to_tdfile, **kw_csv)
                """
                if path_to_tdfile.endswith('.gz'): kw_csv.update(**{'compression': 'gzip'})
                elif path_to_tdfile.endswith('.csv'): kw_csv.pop('compression', None)
                if path_to_tdfile.endswith('.ghg'):
                    with zipfile.ZipFile(path_to_tdfile, 'r') as zip_ref:
                        datafile = [zip_ref.read(name) for name in zip_ref.namelist() if name.endswith(".data")][0]
                    datafile = str(datafile, 'utf-8')
                    path_to_tdfile = StringIO(datafile)
                    # DEFAULT_READ_GHG
                    kw_csv.update(DEFAULT_READ_GHG)
                try:
                    df_td = pd.read_csv(path_to_tdfile, **kw_csv)
                except Exception as e:
                    # (EOFError, pd.errors.ParserError, pd.errors.EmptyDataError):
                    try:
                        if verbosity>1: warnings.warn(f'{e}, when opening {path_to_tdfile}, using {kw_csv}. Re-trying using python as engine and ignoring bad lines.')
                        df_td = pd.read_csv(path_to_tdfile, on_bad_lines='warn', engine='python', **kw_csv)
                    except Exception as ee:
                        warnings.warn(f'{ee}, when opening {str(path_to_tdfile)}, using {kw_csv}')
                        continue
                """
                """
                if kw.tname in df_td.columns:
                    try:
                        df_td.loc[:, kw.tname] = pd.to_datetime(df_td.loc[:, kw.tname].astype(str))
                        print(max(df_td[kw.tname].dt.year), max(df_td[kw.tname]), min(df_td[kw.tname].dt.year))
                        assert max(df_td[kw.tname].dt.year) > 1990 and min(df_td[kw.tname].dt.year) > 1990
                    except:
                        df_td.rename({kw.tname+'_orig': kw.tname})
                """
                if kw.datefomatfrom == 'drop':
                    df_td = df_td.rename({kw.tname: kw.tname+'_orig'})
                
                if kw.tname not in df_td.columns or kw.datefomatfrom == 'drop':
                    if "date" in df_td.columns and "time" in df_td.columns:
                        df_td[kw.tname] = pd.to_datetime(
                            df_td.date + " " + df_td.time, format='%Y-%m-%d %H:%M')
                    else:
                        df_td[kw.tname] = pd.to_datetime(
                            td, format=kw.date_format) - datetime.timedelta(seconds=kw.dt) * (len(df_td)-1 + -1*df_td.index)
                            #td, format=kw.date_format) + datetime.timedelta(seconds=kw.dt) * (df_td.index)
                        df_td[kw.tname] = df_td[kw.tname].dt.strftime(
                            kw.datefomatto)
                else:
                    try:
                        if is_numeric_dtype(df_td[kw.tname]):
                            df_td.loc[:, kw.tname] = df_td.loc[:, kw.tname].apply(lambda e: pd.to_datetime('%.2f' % e, format=kw.datefomatfrom).strftime(kw.datefomatto))
                        elif is_object_dtype(df_td[kw.tname]):
                            df_td.loc[:, kw.tname] = df_td.loc[:, kw.tname].apply(lambda e: pd.to_datetime(e).strftime(kw.datefomatto))
                        else:
                            df_td.loc[:, kw.tname] = pd.to_datetime(df_td[kw.tname], format=kw.datefomatfrom).strftime(kw.datefomatto)
                    except:
                        warnings.warn(f'error when converting {kw.tname} from {kw.datefomatfrom} to {kw.datefomatto}.')
                        continue
                
                df_td['file'] = td
                #df_site = df_site.append(df_td)
                df_site = pd.concat([df_site, df_td], ignore_index=True).reset_index(drop=True)
        
        if df_td.empty == False:
            break
        
    #print('df_td.empty ', df_td.empty)
    if onlynumeric:
        valcols = [i for i in df_site.columns if i.lower() not in [kw.tname.lower(), 'file']]
        _bf = df_site.dtypes
        #df_site.loc[:, valcols] = df_site.loc[:, valcols].apply(pd.to_numeric, errors='coerce')
        df_site[valcols] = df_site[valcols].apply(pd.to_numeric, errors='coerce')
        _af = df_site.dtypes
        if verbosity>1:
            _bfaf = []
            for (k, b) in _bf.items():
                if b!=_af[k]:
                    _nonnum = [s for s in np.unique(df_site[k].apply(lambda s: str(s) if re.findall('[A-z/]+', str(s)) else '')) if s]
                    _bfaf += ['{}, changed from {} to {}. ({})'.format(k, b, _af[k], ', '.join(_nonnum) if _nonnum else 'All numeric')]
            if _bfaf:
                warnings.warn(', '.join(_bfaf))
    
    #if kw_fmt:
    df_site = structuredDataFrame.format(df_site, **kw_['FMT_DATA'])

    if fill:
        if lookup:
            minmax = [min(lookup), max(lookup)]
        else:
            minmax = [np.nanmin(df_site[kw.tname]),
                      np.nanmax(df_site[kw.tname])]
        df_site = df_site.set_index(kw.tname).join(pd.DataFrame({kw.tname: pd.date_range(*minmax, freq=str(kw.dt*1000) + 'ms')}).set_index(kw.tname),
                how='outer').ffill().reset_index()
        #if 'co2' in df_site.columns and (abs(np.max(df_site.co2)) < 1000) and (abs(np.min(df_site.co2)) < 1000):
        #    df_site.loc[:, "co2"] = df_site.loc[:, "co2"] * 1000  # mmol/m3 -> μmol/m3
    
    if kw.id is not None:
        return {kw.id: structuredDataFrame(df_site, dt=kw.dt)}
    else:
        return structuredDataFrame(df_site, dt=kw.dt)


def loaddatawithbuffer(d0, d1=None, freq=None, buffer=None, 
                       tname="TIMESTAMP", f_freq=30, **kwargs):
    if isinstance(d0, (pd.DatetimeIndex, list, set, tuple)) and d1 is None:
        d0, d1 = [np.nanmin(d0), np.nanmax(d0)]
    
    if buffer == None:
        datarange = [pd.date_range(start=d0, end=d1, freq=f"{f_freq}min")[:-1] + pd.Timedelta(freq)]
    else:
        # buffer align with file frequency (e.g. 30 min)
        freqno = int(re.match(r"\d*", f"{f_freq}min")[0])
        bufi = np.ceil(buffer / (freqno*60)) * freqno
        datarange = [
            pd.date_range(
                start=pd.to_datetime(d0) - pd.Timedelta(bufi, unit='min'),
                end=pd.to_datetime(d1) + pd.Timedelta(bufi, unit='min'),
                freq=freq)[:-1] + pd.Timedelta(freq)]
                
    if not datarange:
        return pd.DataFrame()
    
    data = structuredDataFrame(lookup=datarange, **kwargs)
    if data == None or data.data.empty:
        return data.data
    data.data[tname] = pd.to_datetime(data.data[tname])
    
    if buffer:
        d0 = pd.to_datetime(d0) - pd.Timedelta(buffer*1.1, unit='s')
        d1 = pd.to_datetime(d1) + pd.Timedelta(buffer*1.1, unit='s')
        data.filter({tname: (d0, d1)})

    # garantee all data points, if any valid time, else empty dataframe
    if np.sum(np.isnat(data.data.TIMESTAMP)==False):
        #data.data = pd.merge(pd.DataFrame({tname: pd.date_range(*nanminmax(data.data.TIMESTAMP), freq="0.05S")}),
        #                    data.data,
        #                    on=tname, how='outer').reset_index(drop=True)
        return data.data
    else:
        pd.DataFrame()





# built-in modules

# 3rd party modules
import numpy as np
import logging

# project modules

logger = logging.getLogger('corrections')


##########################################
###     SPECTRA                            
##########################################


##########################################
###     DESPIKING                            
##########################################


def mauder2013(x, q=7):
    # it does not do the check for n consecutive spikes 
    x = np.array(x)
    x_med = np.nanmedian(x)
    mad = np.nanmedian(np.abs(x - x_med))
    bounds = (x_med - (q * mad) / 0.6745, x_med + (q * mad) / 0.6745)
    #print("median", x_med, "mad", mad, "bounds", bounds)
    x[x < min(bounds)] = np.nan
    x[x > max(bounds)] = np.nan

    #if fill is not None:
    #    x = fill(pd.Series(x) if fill in (pd.Series.ffill, pd.Series.interpolate) else x)
    return x

def __despike__(X, method=mauder2013):
    N = len(X)
    X = method(X)
    Xna = np.isnan(X)
    try:
        X = np.interp(np.linspace(0, 1, N), 
                        np.linspace(0, 1, N)[Xna == False],
                X[Xna==False])
    except Exception as e:
        logger.error(f"UserWarning: {str(e)}")
    return X


"""
This script is a key part of the following publications:
    - Herig Coimbra, Pedro Henrique and Loubet, Benjamin and Laurent, Olivier and Mauder, Matthias and Heinesch, Bernard and 
    Bitton, Jonathan and Delpierre, Nicolas and Depuydt, Jérémie and Buysse, Pauline, Improvement of Co2 Flux Quality Through 
    Wavelet-Based Eddy Covariance: A New Method for Partitioning Respiration and Photosynthesis. 
    Available at SSRN: https://ssrn.com/abstract=4642939 or http://dx.doi.org/10.2139/ssrn.4642939

The main function is:  
- run_wt
    function: (1) gets data, (2) performs wavelet transform, (3) cross calculate variables, (4) averages by 30 minutes, (5) saves 
    call: run_wt()
    Input:
        a: 
    Return:
        b: 

- conditional_sampling
    function: split an array (n dimensions) into 4 arrays based on signal (+ or -) of itself and 2nd array 
    call: conditional_sampling()
    Input:
        args: arrays to be used as filter 
    Return:
        b: 

- universal_wt
    function: call any wavelet transform
    call: universal_wt()
    Input:
        a: 
    Return:
        b: 
"""

# built-in modules
import os
import re
import warnings
import logging
import copy
import time

# 3rd party modules
from functools import reduce
import numpy as np
import pandas as pd
import itertools
from scipy.optimize import curve_fit
import pywt
try: 
    import pycwt
except ImportError as e:
    pycwt = None
    pass
try: 
    import fcwt
except ImportError as e:
    fcwt = None
    pass

# project modules
from . import commons as hc24
from .commons import j2sj
from . import wavelet_misc as wlmisc

logger = logging.getLogger('wvlt')

def formula_to_vars(formula):
    res = hc24.structuredData()
    res.formula = formula
    res.allvars = formula.replace('|', '*').split('*')
    res.uniquevars = list(set(res.allvars))
    res.xy = formula.split('|')[0].split('*')
    res.condsamp_pair = [v.split('*') for v in formula.split('|')[1:]]
    res.condsamp_flat = [c for cs in res.condsamp_pair for c in cs]
    return res

def __cwt__(input, fs, f0, f1, fn, nthreads=1, scaling="log", fast=False, norm=True, Morlet=6.0):
    """
    function: performs Continuous Wavelet Transform
    call: __cwt__()
    Input:
        a: 
    Return:
        b: 
    """

    #check if input is array and not matrix
    if input.ndim > 1:
        raise ValueError("Input must be a vector")

    #check if input is single precision and change to single precision if not
    if input.dtype != 'single':
        input = input.astype('single')

    morl = fcwt.Morlet(Morlet) #use Morlet wavelet with a wavelet-parameter

    #Generate scales

    if scaling == "lin":
        scales = fcwt.Scales(morl,fcwt.FCWT_LINFREQS,fs,f0,f1,fn)
    elif scaling == "log":
        scales = fcwt.Scales(morl,fcwt.FCWT_LOGSCALES,fs,f0,f1,fn)
    else:
        scales = fcwt.Scales(morl,fcwt.FCWT_LOGSCALES,fs,f0,f1,fn)

    _fcwt = fcwt.FCWT(morl, int(nthreads), fast, norm)

    output = np.zeros((fn,input.size), dtype='csingle')
    freqs = np.zeros((fn), dtype='single')
    
    _fcwt.cwt(input,scales,output)
    scales.getFrequencies(freqs)

    return freqs, output


def __icwt__(W, sj, dt, dj, Cd=None, psi=None, wavelet=None):
    """
    function: performs Inverse Continuous Wavelet Transform
    call: __icwt__()
    Input:
        W: (cross-)spectra
        sj: scales
        dt: sampling rate
        dj: frequency resolution
        Cd, psi: wavelet-specific coefficients
        wavelet: mother wavelet (w/ cdelta and psi(0) callables). Ignored if Cd and psi are given.
    Return:
        x: array
    """
    if wavelet is None: wavelet = pycwt.wavelet.Morlet(6)
    if isinstance(wavelet, str): wavelet = hc24.__wavemother_str_pycwt__(wavelet)
    if Cd is None: Cd = wavelet.cdelta
    if psi is None: psi = wavelet.psi(0)
        
    a, b = W.shape
    c = sj.size
    if a == c:
        sj_ = (np.ones([b, 1]) * sj).transpose()
    elif b == c:
        sj_ = np.ones([a, 1]) * sj
    
    x = (W.real / (sj_ ** .5)) * ((dj * dt ** .5) / (Cd * psi))
    return x

   
def __dwt__(*args, level=None, wavelet="db6"):
    """
    function: performs Discrete Wavelet Transform
    call: __dwt__()
    Input:
        *args: arrays (1D) to be transformed
        level: maximum scale (power of 2)
        wavelet: mother wavelet (comprehensible to pywt)
    Return:
        Ws: list of 2D arrays
    """
    Ws = []
    for X in args:
        Ws += [pywt.wavedec(X, wavelet, level=level)]
    level = len(Ws[-1])-1
    return Ws


def __idwt__(*args, N, level=None, wavelet="db6"):
    """
    function: performs Inverse Discrete Wavelet Transform
    call: __idwt__()
    Input:
        *args: 2D arrays contianing wavelet coefficient
        N: data lenght
        level: maximum scale (power of 2)
        wavelet: mother wavelet (comprehensible to pywt)
    Return:
        Ws: list of 2D arrays
        level: maximum scale (power of 2)
    """
    #assert sum([s==level for s in W.shape]), "Coefficients don't have the same size as level."
    def wrcoef(N, coef_type, coeffs, wavename, level):
        a, ds = coeffs[0], list(reversed(coeffs[1:]))

        if coef_type == 'a':
            return pywt.upcoef('a', a, wavename, level=level, take=N)  # [:N]
        elif coef_type == 'd':
            return pywt.upcoef('d', ds[level-1], wavename, level=level, take=N)  # [:N]
        else:
            raise ValueError("Invalid coefficient type: {}".format(coef_type))
    
    Ys = []
    for W in args:
        A1 = wrcoef(N, 'a', W, wavelet, level)
        D1 = [wrcoef(N, 'd', W, wavelet, i) for i in range(1, level+1)]
        Ys += [np.array(D1 + [A1])]
    return Ys, level


def universal_wt(signal, method='dwt', fs=20, f0=1/(3*60*60), f1=10, fn=180, 
                 dj=1/12, inv=True, **kwargs):
    """
    function: performs Continuous Wavelet Transform
    call: universal_wt()
    Input:
        signal: 1D array
        method: 'dwt', 'cwt', 'fcwt' (cwt but uses fast algorithm)
        fs: sampling rate (Hz)
        f0: highest scale (becomes level for DWT)
        f1: lowest scale (2x sampling rate)
        fn: number of scales (only used for CWT)
        dj: frequency resolution (only used for CWT)
        inv: . Default is True
        **kwargs: keyword arguments sent to wavelet transform and inverse functions 
    Return:
        wave: 2D array
        sj: scales 
    """
    assert method in [
        'dwt', 'cwt', 'fcwt'], "Method not found. Available methods are: dwt, cwt, fcwt"
       
    if method == 'fcwt':
        if fcwt is not None:
            """Run Continuous Wavelet Transform, using fast algorithm"""
            _l, wave = __cwt__(signal, fs, f0, f1, fn, **kwargs)
            sj = np.log2(fs/_l)
            if inv:
                wave = __icwt__(wave, sj=sj, dt=fs, dj=dj, **kwargs, 
                            wavelet=pycwt.wavelet.Morlet(6))
            sj = list(sj)
        else:
            logger.warning('UserWarning: Fast continuous wavelet transform (fcwt) not found. Running slow version.')
            method = 'cwt'
    
    elif method == 'cwt':
        if pycwt is not None:
            """Run Continuous Wavelet Transform"""
            wave, sj, _, _, _, _ = pycwt.cwt(
                signal, dt=1/fs, s0=2/fs, dj=dj, J=fn-1, **kwargs)
            sj = np.log2(sj*fs)
            if inv:
                wave = __icwt__(wave, sj=sj, dt=fs**-1, dj=dj, **kwargs)
            sj = list(sj)
        else:
            logger.warning('UserWarning: Continuous wavelet transform (cwt) not found. Running discrete version.')
            method = 'dwt'
    
    elif method== "dwt":
        """Run Discrete Wavelet Transform"""
        lvl = kwargs.pop('level', int(np.ceil(np.log2(fs/f0))))
        # _l if s0*2^j; fs*2**(-_l) if Hz; (1/fs)*2**_l if sec.
        sj = [_l for _l in np.arange(1, lvl+2, 1)]
        waves = __dwt__(signal, level=lvl, **kwargs)
        if inv:
            N = np.array(signal).shape[-1]
            waves = __idwt__(*waves, N=N, level=lvl, **kwargs)
        wave = waves[0][0]
    return wave.real, sj


def conditional_sampling(Y12, *args, names=['xy', 'a'], label={1: "+", -1: "-"}, false=0):
    # label can also be {1: "+", -1: "-", 0: "·"}
    # guarantee names are enough to name all arguments
    nargs = len(args)
    if nargs < len(names): names = names[:nargs]
    if nargs > len(names): names = names + ['b']* (nargs-len(names))
    # [Y12] + list(args) 
    YS = list(args)
    Ys = {}

    # run for all unique combinations of + and - for groups of size n
    # (e.g., n=2: ++, +-, -+, --, n=3 : +++, ++-, ...)
    for co in set(itertools.combinations(list(label.keys())*nargs, nargs)):
        sign = ''.join([label[c] for c in co])
        name = ''.join([c for cs in zip(names, sign) for c in cs])
        Ys[name] = Y12
        # condition by sign
        for i, c in enumerate(co):
            if c: xy = 1 * (c*YS[i] > 0)
            else: xy = 1 * (YS[i] == 0)
            #xy[xy==0] = false
            xy = np.where(xy == 0, false, xy)
            Ys[name] = Ys[name] * xy
    return Ys


def integrate_cospectra(root, f0, pattern='', dst_path=None):
    if isinstance(root, str):
        saved_files = {}
        for name in os.listdir(root):
            dateparts = re.findall(pattern, name, flags=re.IGNORECASE)
            if len(dateparts) == 1:
                saved_files[dateparts[0]] = os.path.join(root, name)

        def __read__(date, path):
            r = pd.read_csv(path, skiprows=11, sep=',')
            if 'natural_frequency' not in r.columns: 
                logger.warn(f'Skipping spectral file. Natural frequency column not found ({path}).')
                return pd.DataFrame()
            if r.natural_frequency.dtype != float: print(date, r.natural_frequency.dtype)
            r['TIMESTAMP'] = pd.to_datetime(date, format='%Y%m%d%H%M')
            return r

        data = pd.concat([__read__(k, v) for k, v in saved_files.items()])
    else:
        data = root
    
    data0 = data[(np.isnan(data['natural_frequency'])==False) * (data['natural_frequency'] >= f0)].groupby(['variable', 'TIMESTAMP'])['value'].agg(np.nansum).reset_index(drop=False)
    data1 = data[np.isnan(data['natural_frequency'])].drop('natural_frequency', axis=1)

    datai = pd.concat([data1[np.isin(data1['variable'], data0['variable'].unique())==False], data0]).drop_duplicates()
    datai = datai.pivot_table('value', 'TIMESTAMP', 'variable').reset_index(drop=False)
    
    if dst_path: datai.to_file(dst_path, index=False)
    else: return datai
    return

    
def decompose_data(data, variables=['w', 'co2'], dt=0.05, method='dwt', wt_kwargs={}, nan_tolerance=.3, verbosity=1):
    """    Calculate data decomposed with wavelet transform
    """
    assert method in [
        'dwt', 'cwt', 'fcwt'], "Method not found. Available methods are: dwt, cwt, fcwt"
    
    φ = {}
    # run by couple of variables (e.g.: co2*w -> mean(co2'w'))
    info_t_startvarloop = time.time()
    
    # run wavelet transform
    for v in variables:
        if v not in φ.keys():
            signal = np.array(data[v])
            signan = np.isnan(signal)
            N = len(signal)
            Nnan = np.sum(signan)
            if Nnan:
                if (nan_tolerance > 1 and Nnan > nan_tolerance) or (Nnan > nan_tolerance * N):
                    logger.warning(
                        f"UserWarning: Too much nans ({np.sum(signan)}, {np.round(100*np.sum(signan)/len(signal), 1)}%) in {data.TIMESTAMP.head(-1)[0]}.")
            if Nnan and Nnan < N:
                signal = np.interp(np.linspace(0, 1, N), 
                        np.linspace(0, 1, N)[signan == False],
                        signal[signan==False])
            φ[v], sj = universal_wt(signal, method, **wt_kwargs, inv=True)
            N = len(signal)
        

    logger.debug(f'\t\tDecompose all variables took {round(time.time() - info_t_startvarloop)} s (run_wt).')
        
    φs_names = []
    for n in φ.keys():
        if φ[n].shape[0] > 1:
            for l in sj: # use '+ ['']' if __integrate_decomposedarrays_in_dictionary__
                if l: φs_names += [f'{n}_{l}'] 
                else: φs_names += [n] 
        else: φs_names += [n]
    logger.debug(f'\t\tTransform 2D arrays to DataFrame with columns `{"`; `".join(φs_names)}`.')
    logger.debug(f'\t\t{[np.array(v).shape for v in list(φ.values())]}.')
    __temp__ = hc24.matrixtotimetable(np.array(data.TIMESTAMP),
                                      np.concatenate(list(φ.values()), axis=0), columns=φs_names)
    
    __temp__.set_index('TIMESTAMP', inplace=True)
    __temp__.columns = pd.MultiIndex.from_tuples([tuple(c.split('_')) for c in __temp__.columns])
    __temp__ = __temp__.stack(1).reset_index(1).rename(columns={"level_1": "natural_frequency"}).reset_index(drop=False)

    #pattern = re.compile(r"^(?P<variable>.+?)_?(?P<natural_frequency>(?<=_)\d+)?$")
    #__temp__ = __temp__.melt(['TIMESTAMP'] + φ.keys())
    #__temp__ = pd.concat([__temp__.pop('variable').str.extract(pattern, expand=True), __temp__], axis=1)
    __temp__['natural_frequency'] = __temp__['natural_frequency'].apply(lambda j: 1/j2sj(j, 1/dt) if j else np.nan)
    return __temp__


def calculate_product(data, formula='w*co2|w*h2o', verbosity=1):
    """    Calculate (conditioned) product
    """
    var_ = formula_to_vars(formula)

    for ci, c in enumerate(var_.xy): Y12 = Y12 * np.array(data[c]).conjugate() if ci else np.array(data[c])
    φs = {''.join(var_.xy): Y12}

    # conditional sampling
    names = [''.join(var_.xy)] + [''.join(cs) for cs in var_.condsamp_pair]
    φcs = []
    if var_.condsamp_pair: 
        φcs += [Y12]
    for cs in var_.condsamp_pair:
        for ci, c in enumerate(cs): φc0 = φc0 * np.array(data[c]).conjugate() if ci else np.array(data[c])
        φcs += [φc0]
    φc = conditional_sampling(Y12, *φcs, names=names, label={1:"+", -1: "-"}) if φcs else {}
    φs.update(φc)

    data = pd.concat([data, pd.DataFrame(φs)], axis=1)
    return data


def loop_variables(data, varstorun, dt=0.05, wt_kwargs={}, 
                   method="dwt", nan_tolerance=.3, averaging=30, verbosity=1):
    """
    averaging & integrating at the same unit

    fs = 20, f0 = 1/(3*60*60), f1 = 10, fn = 100, agg_avg = 1, 
    suffix = "", wavelet = pycwt.wavelet.MexicanHat(),
    **kwargs):
    """
    wt = {'dt': dt, 'method': method, 'wt_kwargs': wt_kwargs, 'nan_tolerance': nan_tolerance}
    ## START HERE FUNCTION RUN_COV

    # ensure time is time
    data.TIMESTAMP = pd.to_datetime(data.TIMESTAMP)
    
    # ensure continuity
    info_t_startdatacontinuity = time.time()
    data = pd.merge(pd.DataFrame({"TIMESTAMP": pd.date_range(np.nanmin(data.TIMESTAMP), np.nanmax(data.TIMESTAMP), freq=f"{dt*1000}ms")}),
                        data, on="TIMESTAMP", how='outer').reset_index(drop=True)
    timestamp0 = data['TIMESTAMP'].copy()
    timestamp = timestamp0.copy()
    logger.debug(f'\tForce data to be continuous took {round(time.time() - info_t_startdatacontinuity)} s (run_wt).')
    
    assert len(varstorun), 'Empty list of covariances to run. Check available variables and covariances to be performed.'
    allvars = []
    for v in varstorun:
        allvars += formula_to_vars(v).uniquevars

    allvars = list(set(allvars))
    this_data = data[['TIMESTAMP'] + allvars].copy()

    if method == 'cov':
        for c in allvars:
            this_data[c] = this_data[c] - this_data[c].groupby(this_data.TIMESTAMP.dt.floor(f'{averaging}min')).transform(np.nanmean)
        
        this_data = pd.concat(
            [(calculate_product(this_data, formula=f)
                .drop(columns=formula_to_vars(f).uniquevars if i == 0 else ['TIMESTAMP'] + formula_to_vars(f).uniquevars))
                for i, f in enumerate(varstorun)], axis=1)
        #this_data = calculate_product(this_data, formula=var_.formula)
        return this_data

    else:
        info_t_startdecomposedata = time.time()
        data_decomp = decompose_data(this_data, variables=allvars, **wt)
        logger.debug(f'\tDecompose data took {round(time.time() - info_t_startdecomposedata)} s (run_wt).')

        info_t_startconcatenatedata = time.time()

        data_decomp = pd.concat(
            [(calculate_product(data_decomp, formula=f)
                .drop(columns=formula_to_vars(f).uniquevars if i == 0 else ['TIMESTAMP', 'natural_frequency'] + formula_to_vars(f).uniquevars))
                for i, f in enumerate(varstorun)], axis=1)
        return data_decomp

    #[os.remove(d) for a in avg_ for d in dat_fullspectra[a] if os.path.exists(d)]
    #if os.path.exists(curoutpath_inprog): os.remove(curoutpath_inprog)
    #prev_print = '\x1B[2A\r' + f' {date} {len(yl)} files {int(100*ymd_i/len(ymd))} % ({time.strftime("%d.%m.%y %H:%M:%S")})' + '\n'
        
    #logger.info(f'\tFinish {date} took {round(time.time() - info_t_startdateloop)} s, yielded {len(yl)} files (run_wt). Progress: {len(yl)} {int(100*ymd_i/len(ymd))} %')
    data = pd.concat([data_decomp, this_data], axis=0)
    logger.debug(f'\tConcatenate data took {round(time.time() - info_t_startconcatenatedata)} s (run_wt).')
    return data

def load_data_and_loop(ymd, raw_kwargs, output_path=None, verbosity=1,
                  overwrite=False, processing_time_duration="1D", 
                  internal_averaging=None, dt=0.05, wt_kwargs={}, 
                  method="dwt", averaging=30, **kwargs):
    if isinstance(averaging, (list, tuple)): averaging = averaging[-1]
    if internal_averaging is None: internal_averaging = averaging
    fulldata = pd.DataFrame()

    info_t_start = time.time()
    logger.info('Entered wavelet code (run_wt).')

    if verbosity: print(f'\nRUNNING WAVELET TRASNFORM ({method})')

    suffix = '_cov' if method == 'cov' else '_full_cospectra'

    _, _, _f = ymd
    ymd = hc24.list_time_in_period(*ymd, processing_time_duration, include='both')
    
    if method == 'cov':
        buffer = 0
    elif method in ['dwt']:
        buffer = wlmisc.bufferforfrequency_dwt(n_=_f, **wt_kwargs)/2
    else:
        buffer = wlmisc.bufferforfrequency(wt_kwargs.get("f0", 1/(3*60*60))) / 2
    logger.debug(f"Buffer (s): {buffer}.")
        
    logger.info(f'Start date loop at {round(time.time() - info_t_start)} s (run_wt).')
    
    # Skip two line
    prev_print = '\n'
    for ymd_i, yl in enumerate(ymd):
        legitimate_to_write = 0
        info_t_startdateloop = time.time()

        date = re.sub('[-: ]', '', str(yl[0]))
        if processing_time_duration.endswith("D"): date = date[:8]
        if processing_time_duration.endswith("H") or processing_time_duration.endswith("Min"): date = date[:12]
        
        print(prev_print, date, 'reading', ' '*10, sep=' ', end='\n')
        prev_print = '\x1B[1A\r'

        if output_path:
            # recheck if files exist and overwrite option
            # doesn't save time (maybe only save 5min)
            if not overwrite and os.path.exists(output_path.format(suffix, date)):
                logger.warning("UserWarning: Skipping, file already exists ({}).".format(date))
                continue
            
            if all([os.path.exists(output_path.format(suffix, _yl.strftime('%Y%m%d%H%M'))) for _yl in yl[:-1]]): 
                logger.warning("UserWarning: Skipping, file already exists ({}).".format(date))
                continue
            elif any([os.path.exists(output_path.format(suffix, _yl.strftime('%Y%m%d%H%M'))) for _yl in yl[:-1]]):
                logger.warning("UserWarning: Continuing but some files already exist ({}), others don't ({}).".format(
                    ', '.join([_yl.strftime('%Y%m%d%H%M') for _yl in yl[:-1] if os.path.exists(output_path.format(suffix, _yl.strftime('%Y%m%d%H%M')))]), 
                    ', '.join([_yl.strftime('%Y%m%d%H%M') for _yl in yl[:-1] if not os.path.exists(output_path.format(suffix, _yl.strftime('%Y%m%d%H%M')))]), 
                    ))
                pass
            
            curoutpath_inprog = output_path.format(suffix, str(date), "").rsplit(".", 1)[
                0] + ".inprogress"
            if hc24.checkifinprogress(curoutpath_inprog): continue
        
        # load files
        info_t_startloaddata = time.time()
        data = hc24.loaddatawithbuffer(
            yl, d1=None, freq=_f, buffer=buffer, f_freq=_f, **raw_kwargs)
        if data.empty:
            if verbosity>1: logger.warning("UserWarning: No file was found ({}, path: {}).".format(date, raw_kwargs.get('path', 'default')))
            if output_path and os.path.exists(curoutpath_inprog): os.remove(curoutpath_inprog)
            continue
        logger.info(f'\tLoading data took {round(time.time() - info_t_startloaddata)} s (run_wt).')
        
        # LOOP FOR VARIABLES
        #__save_cospectra__(data, averaging=internal_averaging,
        #                          dt=dt, wt_kwargs=wt_kwargs, method=method, 
        #                          verbosity=verbosity, **kwargs)
        try:
            data = loop_variables(data, averaging=internal_averaging,
                                  dt=dt, wt_kwargs=wt_kwargs, method=method, 
                                  verbosity=verbosity, **kwargs)
            data['TIMESTAMP'] = data['TIMESTAMP'].dt.floor(f'{averaging}min')
            __ID_COLS__ = ['TIMESTAMP'] if method == "cov" else ['TIMESTAMP', 'natural_frequency']
            data = data.groupby(__ID_COLS__).agg(np.nanmean).reset_index(drop=False)
        except Exception as e:
            logger.critical(str(e))
            print(str(e))

        #[os.remove(d) for a in avg_ for d in dat_fullspectra[a] if os.path.exists(d)]
        if output_path and os.path.exists(curoutpath_inprog): os.remove(curoutpath_inprog)
        prev_print = '\x1B[2A\r' + f' {date} {len(yl)} files {int(100*ymd_i/len(ymd))} % ({time.strftime("%d.%m.%y %H:%M:%S")})' + '\n'
            
        logger.info(f'\tFinish {date} took {round(time.time() - info_t_startdateloop)} s, yielded {len(yl)} files (run_wt). Progress: {len(yl)} {int(100*ymd_i/len(ymd))} %')

        logger.debug(f'\tSaving data starting at {round(time.time() - info_t_start)} s (run_wt).')
        
        if method == 'cov':
            fulldata = pd.concat([fulldata, data], axis=0)
        else:
            # FILTER OUT THE BUFFER DATA, KEEP ONLY GOOD DATA 
            #data = data[data.TIMESTAMP>= ]
            data = (data.sort_values(by=['TIMESTAMP', 'natural_frequency'])
                    .melt(['TIMESTAMP', 'natural_frequency']))
            if output_path:
                __save_cospectra__(data, suffix, output_path, **{'method': method, 'averaging': averaging, 'buffer': buffer, 'dt': dt})
            else:
                fulldata = pd.concat([fulldata, data], axis=0)
    
    if output_path and not fulldata.empty:
            fulldata.to_csv(output_path.format(suffix, pd.datetime.now().strftime('%Y%m%dT%H%M%S_%f')), index=False)
    return fulldata


def __save_cospectra__(data, suffix, output_path, **meta):
    info_t_startsaveloop = time.time()
    for __datea__, __tempa__ in data.groupby(data.TIMESTAMP):
        dst_path = output_path.format(suffix, pd.to_datetime(__datea__).strftime('%Y%m%d%H%M'))
        if os.path.exists(dst_path): continue
        use_header = False
        if not os.path.exists(dst_path):
            use_header = True
            header  = "wavelet_based_(co)spectra\n"
            header += f"--------------------------------------------------------------\n"
            header += f"TIMESTAMP_START = {min(__tempa__.TIMESTAMP)}\n"
            header += f"TIMESTAMP_END = {max(__tempa__.TIMESTAMP)}\n"
            header += f"N: {len(__tempa__.TIMESTAMP)}\n"
            header += f"TIME_BUFFER [min] = {meta.get('buffer', '')/60}\n"
            header += f"frequency [Hz]\n"
            header += f"y-axis_->_wavelet_coefficient_*_\n"
            header += f"mother_wavelet -> {meta.get('method', '')}\n"
            header += f"acquisition_frequency [Hz] = {1/meta.get('dt', '')}\n"
            header += f"averaging_interval [Min] = {meta.get('averaging', '')}\n"
            hc24.mkdirs(dst_path)
            with open(dst_path, 'w+') as part: part.write(header)
            legitimate_to_write = 1
            logger.debug(f'\t\tSaving header of DataFrame took {round(time.time() - info_t_startsaveloop)} s (run_wt).')
        
        if not legitimate_to_write: continue
        
        __tempa__.drop('TIMESTAMP', axis=1, inplace=True)
        with open(dst_path, 'a+', newline='') as part: __tempa__.to_file(part, header=use_header, chunksize=500, index=False)
        
        del __tempa__
        
    #arr_slice = np.unique(data.TIMESTAMP, return_index=True)
    #for __datea__ in arr_slice[0]:
    #    dst_path = output_path.format(suffix, pd.to_datetime(__datea__).strftime('%Y%m%d%H%M'))
    #    if os.path.exists(dst_path+'.part'): os.rename(dst_path+'.part', dst_path)
    return


def run_wt(ymd, varstorun, raw_kwargs, output_path, wt_kwargs={}, 
           method="dwt", Cφ=1, nan_tolerance=.3,
           averaging=30, condsamp_flat=[], integrating=30*60, 
           overwrite=False, saveraw=False, processing_time_duration="1D",
           preaverage=None,
           despike=False, denoise=False, noisecolor=1, deadband={}, verbosity=1):
    """
    averaging & integrating at the same unit

    fs = 20, f0 = 1/(3*60*60), f1 = 10, fn = 100, agg_avg = 1, 
    suffix = "", wavelet = pycwt.wavelet.MexicanHat(),
    **kwargs):
    """
    if isinstance(averaging, (list, tuple)): averaging = averaging[-1]

    info_t_start = time.time()
    logger.info('Entered wavelet code (run_wt).')

    assert method in [
        'dwt', 'cwt', 'fcwt'], "Method not found. Available methods are: dwt, cwt, fcwt"
    if verbosity: print(f'\nRUNNING WAVELET TRASNFORM ({method})')
    if False and (method in ["cwt", "fcwt"]):
        if method == "fcwt" or "wavelet" not in wt_kwargs.keys() or wt_kwargs.get("wavelet") in ['morlet', 'Morlet', pycwt.wavelet.Morlet(6)]:
            Cφ = 2.5
        else:
            Cφ = 16.568
    
    dt = 1 / wt_kwargs.get("fs", 20)
    suffix = raw_kwargs['suffix'] if 'suffix' in raw_kwargs.keys() else ''

    _, _, _f = ymd
    ymd = hc24.list_time_in_period(*ymd, processing_time_duration, include='both')
    
    if method in ['dwt']:
        buffer = wlmisc.bufferforfrequency_dwt(
            N=pd.to_timedelta(processing_time_duration)/pd.to_timedelta("1s") * dt**-1,
            n_=_f, **wt_kwargs)/2
    else:
        buffer = wlmisc.bufferforfrequency(wt_kwargs.get("f0", 1/(3*60*60))) / 2
    logger.debug(f"Buffer (s): {buffer}.")
        
    logger.info(f'Start date loop at {round(time.time() - info_t_start)} s (run_wt).')
    
    def __despike__(X):
        N = len(X)
        X = hc24.mauder2013(X)
        Xna = np.isnan(X)
        try:
            X = np.interp(np.linspace(0, 1, N), 
                            np.linspace(0, 1, N)[Xna == False],
                    X[Xna==False])
        except Exception as e:
            logger.error(f"UserWarning: {str(e)}")
        return X
    
    def __fit_whitenoise__(spec, fmax=5, a=noisecolor):
        r = hc24.structuredData()
        freqs = [1/j2sj(j, 1/dt) for j in np.arange(1,len(spec)+1)]
        r.curve_0 = lambda f, b: np.log((f**a)*b)
        specna = np.where(np.isnan(spec[:fmax]) | (spec[:fmax] <= 0), False, True)
        try:
            freqs = np.array(freqs)
            spec  = np.array(spec)
            r.params_0, _ = curve_fit(r.curve_0, freqs[:fmax][specna], np.log(spec[:fmax][specna]), bounds=(0, np.inf))
        except Exception as e:
            logger.error(f"UserWarning: {str(e)}")
            r.params_0 = [0]
        r.fit = np.array([(f**a)*r.params_0[0] for f in freqs])
        return r

    # Skip two line
    prev_print = '\n'
    for ymd_i, yl in enumerate(ymd):
        legitimate_to_write = 0
        info_t_startdateloop = time.time()

        date = re.sub('[-: ]', '', str(yl[0]))
        if processing_time_duration.endswith("D"): date = date[:8]
        if processing_time_duration.endswith("H") or processing_time_duration.endswith("Min"): date = date[:12]
        
        print(prev_print, date, 'reading', ' '*10, sep=' ', end='\n')
        prev_print = '\x1B[1A\r'

        # recheck if files exist and overwrite option
        # doesn't save time (maybe only save 5min)
        if not overwrite and os.path.exists(output_path.format(suffix + "_full_cospectra", date)):
            logger.warning("UserWarning: Skipping, file already exists ({}).".format(date))
            continue
        
        if all([os.path.exists(output_path.format(suffix + "_full_cospectra", _yl.strftime('%Y%m%d%H%M'))) for _yl in yl[:-1]]): 
            logger.warning("UserWarning: Skipping, file already exists ({}).".format(date))
            continue
        elif any([os.path.exists(output_path.format(suffix + "_full_cospectra", _yl.strftime('%Y%m%d%H%M'))) for _yl in yl[:-1]]):
            logger.warning("UserWarning: Continuing but some files already exist ({}), others don't ({}).".format(
                ', '.join([_yl.strftime('%Y%m%d%H%M') for _yl in yl[:-1] if os.path.exists(output_path.format(suffix + "_full_cospectra", _yl.strftime('%Y%m%d%H%M')))]), 
                ', '.join([_yl.strftime('%Y%m%d%H%M') for _yl in yl[:-1] if not os.path.exists(output_path.format(suffix + "_full_cospectra", _yl.strftime('%Y%m%d%H%M')))]), 
                ))
            pass

        """
        if not overwrite:
            avg_ = []
            for a in averaging:
                if not overwrite and os.path.exists(output_path.format(suffix + "_full_cospectra", date, str(a).zfill(2))):
                    avg_ += [a]
            avg_ = list(set(averaging)-set(avg_))
            if not avg_:
                if verbosity > 1: logger.warning("UserWarning: Skipping, file already exists ({}).".format(date))
                continue
        else:
            avg_ = [a for a in averaging]
        """
        
        curoutpath_inprog = output_path.format(suffix, str(date), "").rsplit(".", 1)[
            0] + ".inprogress"
        if hc24.checkifinprogress(curoutpath_inprog): continue
        
        # load files
        # data = get_rawdata.open_flux(lookup=yl, **raw_kwargs).data
        info_t_startloaddata = time.time()
        data = hc24.loaddatawithbuffer(
            yl, d1=None, freq=_f, buffer=buffer, f_freq=_f, **raw_kwargs)
        if data.empty:
            if verbosity>1: logger.warning("UserWarning: No file was found ({}, path: {}).".format(date, raw_kwargs.get('path', 'default')))
            if os.path.exists(curoutpath_inprog): os.remove(curoutpath_inprog)
            continue
        logger.info(f'\tLoading data took {round(time.time() - info_t_startloaddata)} s (run_wt).')
        
        ## START HERE FUNCTION RUN_COV

        # ensure time is time
        data.TIMESTAMP = pd.to_datetime(data.TIMESTAMP)
        
        # ensure continuity
        info_t_startdatacontinuity = time.time()
        data = pd.merge(pd.DataFrame({"TIMESTAMP": pd.date_range(np.nanmin(data.TIMESTAMP), np.nanmax(data.TIMESTAMP), freq=f"{dt*1000}ms")}),
                            data, on="TIMESTAMP", how='outer').reset_index(drop=True)
        timestamp0 = data['TIMESTAMP'].copy()
        timestamp = timestamp0.copy()
        logger.debug(f'\tForce data to be continuous took {round(time.time() - info_t_startdatacontinuity)} s (run_wt).')

        # main run
        # . collect all wavelet transforms
        # . calculate covariance
        # . conditional sampling (optional)
        # . save in dataframe and .csv
        φ = {}
        φ0 = {}
        σ = {}
        #ζ = {}
        ζb = {}
        S = {}
        μ = {}
        dat_fullspectra_file = []  # {a: [] for a in avg_}
        dat_fullspectra = []  # {a: [] for a in avg_}
        dat_fluxresult = []  # {a: [] for a in avg_}

        print(prev_print, date, 'decomposing', ' '*10, sep=' ', end='\n')
        # run by couple of variables (e.g.: co2*w -> mean(co2'w')) 
        logger.info(f'\tStarting variables loop at {round(time.time() - info_t_start)} s (run_wt).')
        info_t_dataframe     = 0 
        info_t_varstorunloop = 0
        info_c_varstorunloop = 0
        try:
            assert len(varstorun), 'Empty list of covariances to run. Check available variables and covariances to be performed.'
            for thisvar_i, thisvar in enumerate(varstorun):
                info_t_startvarloop = time.time()
                
                allvars = thisvar.replace('|', '*').split('*')
                xy = thisvar.split('|')[0].split('*')
                condsamp_pair = [v.split('*') for v in thisvar.split('|')[1:]]
                condsamp_flat = [c for cs in condsamp_pair for c in cs]
                #for xy, condsamp_flat in [(v.split('*')[:2], v.split('*')[2:]) for v in varstorun]:
                if '|' in thisvar: logger.debug(f"\t\tCall {thisvar} ({'*'.join(xy)} conditioned by {thisvar.split('|', 1)[1]}).")
                else: logger.debug(f"\t\tCall {thisvar}.")
                
                # run wavelet transform
                for v in xy + condsamp_flat:
                    if v not in φ.keys():
                        info_c_varstorunloop += 1
                        info_t_startdecomposition_v = time.time()  
                        signal = np.array(data[v])
                        signan = np.isnan(signal)
                        N = len(signal)
                        Nnan = np.sum(signan)
                        if Nnan:
                            if (nan_tolerance > 1 and Nnan > nan_tolerance) or (Nnan > nan_tolerance * N):
                                logger.warning(
                                    f"UserWarning: Too much nans ({np.sum(signan)}, {np.round(100*np.sum(signan)/len(signal), 1)}%) in {date}.")
                        if Nnan and Nnan < N:
                            signal = np.interp(np.linspace(0, 1, N), 
                                    np.linspace(0, 1, N)[signan == False],
                                    signal[signan==False])
                        φ[v], sj = universal_wt(signal, method, **wt_kwargs, inv=True)
                        levels_to_integrate = [dt*2**l < integrating for l in sj]
                        levels_notto_integrate = [l==False for l in levels_to_integrate]

                        # Strip buffer data
                        _isinfinalrange = (timestamp0 >= min(yl)) & (timestamp0 < max(yl))
                        signal = signal[_isinfinalrange.to_list()]
                        signan = signan[_isinfinalrange.to_list()]
                        φ[v] = φ[v][:, _isinfinalrange.to_list()]
                        N = len(signal)
                        S[v] = signal

                        # apply despiking (Mauder et al.)
                        if despike:
                            φ[v] = np.apply_along_axis(__despike__, 1, φ[v])
                        
                        φ0[v] = copy.deepcopy(φ[v])
                        μ[v] = signan *1
                        μ[v+'D'] = signan *1
                        μ[v+'T'] = signan *1
                        
                        """
                        # apply Allan deviation (ad-hoc)
                        frequencies = [dt*2**l for l in sj]
                        # read allan deviation file
                        # interpolate and find frequencies

                        allan_dev = {'co2': np.array([0.014, 0.014, 0.014, 0.014, 0.013, 0.01,  0.007, 0.005, 0.004, 0.004, 0.003, 0.002, 0.002, 0.001, 0.002, 0.003, 0.003, np.inf]),
                                     'co': np.array([3.095, 3.095, 3.095, 3.095, 2.853, 2.206, 1.393, 0.998, 0.674, 0.544, 0.418, 0.265, 0.168, 0.124, 0.103, 0.067, 0.086, np.inf]),
                                     'ch4': np.array([0.089, 0.089, 0.089, 0.089, 0.083, 0.065, 0.043, 0.034, 0.027, 0.023, 0.018, 0.013, 0.01,  0.011, 0.018, 0.027, 0.041, np.inf])}
                        allan_dev = {k: np.expand_dims(v, 1) for k, v in allan_dev.items()}

                        #filter values
                        if (N > 0) and (v.lower() in allan_dev.keys()):
                            # φ[v] needs to be (n, 17)
                            φ0[v] = np.where((abs(φ[v]) < np.sqrt(allan_dev[v])), 0, φ0[v])
                        
                        #if still some values
                        ##fit theorical (co)spectrum
                        ##fill spectra using theorical curve
                        """

                        # for file length, update later
                        σ[v]  = []
                        σ[v+'D'] = []
                        σ[v+'T'] = []
                        #ζ[v]  = []
                        ζb[v] = []

                        #arr_slice = np.unique(data[_isinfinalrange].TIMESTAMP.dt.floor(_f), return_index=True)
                        arr_slice = np.unique(timestamp0[_isinfinalrange].dt.floor(_f), return_index=True)
                        for this_group, this_data in list(zip(arr_slice[0], np.split(timestamp0[_isinfinalrange], arr_slice[1][1:]))):
                            #for this_group, this_data in data.groupby(data[_isinfinalrange].TIMESTAMP.dt.floor(_f)):
                            # Everytime creating working with the whole array, maybe faster to work on a split of the array (redability)
                            # List of T/F to select values that are in the current AVERAGING period
                            #this_hh = np.isin(data[_isinfinalrange].index, this_data.index)#np.where(data.TIMESTAMP > this_yl) & (data.TIMESTAMP <= (this_yl + pd.Timedelta(_f))), True, False)
                            this_hh = np.isin(timestamp0[_isinfinalrange].index, this_data.index)
                            this_Sv = np.array(S[v])
                            this_trend = np.nansum(np.array(φ[v])[levels_notto_integrate], axis=0)
                            this_N = len(this_Sv[this_hh])
                            σv = this_Sv - np.nanmean(this_Sv[this_hh])
                            σ[v]  += list(σv[this_hh])
                            σ[v+'D']  += list(this_Sv[this_hh] - this_trend[this_hh])
                            σ[v+'T']  += list(this_trend[this_hh])
                                
                            if v in deadband.keys():
                                this_hh = this_hh & (abs(σv) > abs(deadband[v])) #(σv > deadband[v])
                                φ0[v] = np.where((abs(σv) < abs(deadband[v])) * this_hh.reshape(1, -1), 0, φ0[v])

                            if denoise and (this_data.empty==False):
                                this_φv = np.array(φ[v])[:, this_hh]
                                σ2_v = np.nanmean(σv**2)
                                freq_noise_fit = 5
                                if dt==0.1: freq_noise_fit = 4
                                # np.nanmedian(abs(this_φv), axis=1)
                                noise_fitted = __fit_whitenoise__(np.nanmedian((this_φv**2)/σ2_v, axis=1), freq_noise_fit)
                                
                                this_noise = abs(np.sqrt(noise_fitted.fit * σ2_v).reshape(-1, 1))

                                #ζ[v]  += [this_noise]*len(this_Sv)
                                ζb[v] += [noise_fitted.params_0[0]]*this_N
                                
                                φ0[v] = np.where((abs(φ0[v]) > abs(this_noise)) * this_hh.reshape(1, -1) * (φ0[v] > 0), φ0[v] - this_noise, φ0[v])
                                φ0[v] = np.where((abs(φ0[v]) > abs(this_noise)) * this_hh.reshape(1, -1) * (φ0[v] < 0), φ0[v] + this_noise, φ0[v])
                                φ0[v] = np.where((abs(φ0[v]) <= abs(this_noise)) * this_hh.reshape(1, -1), 0, φ0[v])
                            else:
                                #ζ[v] += list(np.zeros((this_N, 1)))
                                ζb[v] += list(np.zeros((this_N, 1)) * np.nan)
                            
                        
                        # arrayify
                        φ[v] = np.array(φ[v])
                        φ0[v] = np.array(φ0[v])
                        σ[v] = np.array(σ[v]).reshape(1,-1)
                        σ[v+'D'] = np.array(σ[v+'D']).reshape(1,-1)
                        σ[v+'T'] = np.array(σ[v+'T']).reshape(1,-1)
                        #logger.debug(f'\t\tDecompose all variables took {σ[v].shape} / {σ[v+"D"].shape} s (run_wt).')
                        #ζ[v] = np.array(ζ[v]).T
                        ζb[v] = np.array(ζb[v]).reshape(1,-1)
                        S[v] = np.array(S[v]).reshape(1,-1)
                        #μ[v] = np.array(μ[v]).reshape(1,-1)
                logger.debug(f'\t\tDecompose all variables took {round(time.time() - info_t_startvarloop)} s (run_wt).')
                info_t_varstorunloop += time.time() - info_t_startvarloop

                info_t_startcovariance = time.time()
                # Strip definitively buffer data
                # data = data[_isinfinalrange]

                # calculate covariance
                for ci, c in enumerate(xy): Y12 = Y12 * φ[c].conjugate() if ci else φ[c] * Cφ
                for ci, c in enumerate(xy): Y120 = Y120 * φ0[c].conjugate() if ci else φ0[c] * Cφ
                for ci, c in enumerate(xy): Cov = Cov * σ[c] if ci else copy.deepcopy(σ[c])
                                
                CovT = {}
                for t in ['T', 'D']:
                    for co in set(itertools.combinations(['', t]*len(xy), len(xy))):
                        for ci, c in enumerate(xy): Cov_ = Cov_ * σ[c+co[ci]] if ci else copy.deepcopy(σ[c+co[0]])
                        CovT['/'.join([c+co[ci] for ci, c in enumerate(xy)])] = Cov_
                
                # addictive so it only remains which is > uncertainty
                #for ci, c in enumerate(xy): Y12ζ = Y12 * φ[c].conjugate() if ci else φ[c] * Cφ
                #for ci, c in enumerate(xy): Y12ζ = np.where((abs(φ[c]) > abs(ζ[c])) & (abs(φ[c])>0), 
                #                                            Y12ζ * (φ[c].conjugate() - (ζ[c] * φ[c]/abs(φ[c]))), 
                #                                            0) if ci else φ[c] * Cφ
                #Y12ζ = (abs(Y12) - abs(ζ[xy[-1]])) * Y12/abs(Y12)
                #print(date, f"\t call: {thisvar} ({'*'.join(xy)}", f"conditioned by {thisvar.split('|', 1)[1]})" if '|' in thisvar else ')',
                #    'main', '\t shape:', Y12.shape, '\t N (days):', round(Y12.shape[1] / (24*60*60/dt), 2), '\t buffer (s):', buffer)
                logger.debug(f"\t\tDecomposed covariance shape: {Y12.shape} ({round(Y12.shape[1] / (24*60*60/dt), 2)} days) (run_wt).")
                
                φs = {''.join(xy): Y12, '/'.join(xy)+'_cov': Cov.reshape(1,-1)}
                φs.update({f'{k}_cov': v.reshape(1,-1) for k, v in CovT.items()})
                φs.update({f'{k}_var': (v**2).reshape(1,-1) for k, v in σ.items()})
                μs = {''.join(xy): np.where(np.where(
                    np.array(μ[xy[0]]), 0, 1) * np.where(np.array(μ[xy[1]]), 0, 1), 0, 1).reshape(1,-1)}
                μs.update({'/'.join(xy)+'_cov': μs[''.join(xy)]})
                # Ad-hoc solution
                μs.update({f'{k}_cov': μs[k.replace('D', '').replace('T', '')+'_cov'] for k in CovT.keys()})
                μs.update({f'{k}_var': np.where(np.array(μ[k]), 0, 1).reshape(1,-1) for k in σ.keys()})
                
                logger.debug(f'\t\tCalculate covariances took {round(time.time() - info_t_startcovariance)} s (run_wt).')
                
                #Pre average to avoid (e.g.: 1 min)
                #It has to be divisor of 30 min to preserve 30 min average
                info_t_startpreaveraging = time.time()
                if preaverage:
                    arr_slice = np.unique(timestamp0[_isinfinalrange].dt.floor(preaverage), return_index=True)
                    #print(arr_slice[1][1:])
                    #print(len([np.nanmean(s) for s in np.split(signal, arr_slice[1][1:])]))
                    #print([len(s) for s in [s for s in np.split(signal, arr_slice[1][1:])]])
                    #signal = np.array([np.nanmean(s) for s in np.split(signal, arr_slice[1][1:])])
                    #signan = np.array([np.nanmean(s) for s in np.split(signan, arr_slice[1][1:])])
                    #φ[v] = np.array([np.nanmean(s, axis=0) for s in np.split(φ[v].T, arr_slice[1][1:])]).T
                    
                    φ0 = {k: np.array([np.nanmean(s, axis=0) for s in np.split(v.T, arr_slice[1][1:])]).T for k, v in φ0.items()}
                    φs = {k: np.array([np.nanmean(s, axis=0) for s in np.split(v.T, arr_slice[1][1:])]).T for k, v in φs.items()}
                    logger.debug(f'\tPreaveraging φs took {round(time.time() - info_t_startpreaveraging)} s (run_wt).')
                    μs = {k: np.array([np.nanmean(s, axis=0) for s in np.split(v.T, arr_slice[1][1:])]).T for k, v in μs.items()}
                    logger.debug(f'\tPreaveraging μs took {round(time.time() - info_t_startpreaveraging)} s (run_wt).')
                    ζb = {k: np.array([np.nanmean(s, axis=0) for s in np.split(v.T, arr_slice[1][1:])]).T for k, v in ζb.items()}
                    logger.debug(f'\tPreaveraging ζb took {round(time.time() - info_t_startpreaveraging)} s (run_wt).')
                    Y12 = np.array([np.nanmean(s, axis=0) for s in np.split(Y12.T, arr_slice[1][1:])]).T
                    Y120 = np.array([np.nanmean(s, axis=0) for s in np.split(Y120.T, arr_slice[1][1:])]).T
                    logger.debug(f'\tPreaveraging Y12 took {round(time.time() - info_t_startpreaveraging)} s (run_wt).')
                    timestamp = pd.Series(arr_slice[0])
                else:
                    timestamp = timestamp0[_isinfinalrange].copy()
                logger.debug(f'\tPreaveraging took {round(time.time() - info_t_startpreaveraging)} s (run_wt).')
                
                info_t_startconditionalsampling = time.time()
                # conditional sampling
                names = [''.join(xy)] + [''.join(cs) for cs in condsamp_pair]
                φcs = []
                if condsamp_pair: 
                    φcs += [Y120]#[Y12ζ]
                for cs in condsamp_pair:
                    for ci, c in enumerate(cs): φc0 = φc0 * φ0[c].conjugate() if ci else φ0[c]
                    # if variables to condition are uncertain do not consider them
                    # and put them in uncertain bag (Y12ζ) 
                    #for ci, c in enumerate(cs): φc0  = np.where(abs(φ[c]) > abs(ζ[c]), φc0, 0)
                    #for ci, c in enumerate(cs): Y12ζ = np.where(abs(φ[c]) > abs(ζ[c]), Y12ζ, 0)
                    φcs += [φc0]
                
                if ''.join(xy) in ['wco2', 'co2w', 'wh2o', 'h2ow']:
                    σcec = conditional_sampling(Cov, *[σ['w'], σ['co2'], σ['h2o']], 
                                                names=[''.join(xy), 'w', 'co2', 'h2o'],
                                                label={1: "+", -1: "-"})
                    φs.update(σcec)

                """
                names = [''.join(xy)] + [xy[0]+c for c in condsamp_flat]
                φc = [np.array(φ[xy[0]]) * np.array(φ[c]).conjugate() if xy[0] != c else np.array(φ[c]) for c in condsamp_flat]
                """
                #Y12ζ
                logger.debug(f'\t\tDifference between original and denoised {np.nanmean(abs(Y12 - Y120))} for array w/ shape {Y12.shape} and {Y120.shape} (run_wt).')
                φc = conditional_sampling(Y12, *φcs, names=names) if φcs else {}
                if φc: logger.debug(f'\t\tDifference between original and partitioned {np.nanmean(abs(Y12 - np.nansum(list(φc.values()), 0)))} for sum of {φc.keys()} (run_wt).')
                #φs.update({k.replace("xy", ''.join(xy)).replace('a', ''.join(condsamp_flat)): v for k, v in φc.items()})
                φs.update(φc)
                
                # ζY12 should be:
                # - 0, when  φ = 0
                # - φ, when |ζ|<|φ|
                # - ζ, when |ζ|>|φ|
                ζY12 = Y12 - Y120 #Y12ζ  # np.where(Y12ζ==0, np.max(Y12, ), 0)
                logger.debug(f'\t\tDone normal conditional sampling took {round(time.time() - info_t_startconditionalsampling)} s (run_wt).')
                names0 = [''.join(xy) + '·' + ''.join(condsamp_pair[0])] + [''.join(cs) for cs in condsamp_pair[1:]] if φcs else []
                φc0 = conditional_sampling(ζY12, *φcs[1:], names=names0) if φcs else {}
                #φs.update({k.replace("xy", ''.join(xy)).replace('a', ''.join(condsamp_flat)): v for k, v in φc.items()})
                φs.update(φc0)
                
                if φc and φc0: logger.debug(f'\t\tDifference between original and partitioned {np.nanmean(abs(Y12 - np.nansum(list(dict(φc, **φc0).values()), 0)))} for sum of {φc.keys()} (run_wt).')
                
                # add noise columns
                φs.update({''.join(xy) + '·': ζY12})

                logger.debug(f'\t\tConditional sampling took {round(time.time() - info_t_startconditionalsampling)} s (run_wt).')
                
                # repeats nan flag wo/ considering conditional sampling variables
                μs.update(
                    {k: μs[k if k in μs.keys() else [k_ for k_ in μs.keys() if k.startswith(k_)][0]] for k in φs.keys()})
                logger.debug(f'\t\tAdding nan flag took {round(time.time() - info_t_startconditionalsampling)} s (run_wt).')

                info_t_startdataframe = time.time()
                # array to dataframe for averaging
                
                # Integrate arrays are 2D inside dictionary (needed for concatenate)
                def __integrate_decomposedarrays_in_dictionary__(array_dic):
                    levels_to_integrate = [1/j2sj(l, 1/dt) >= 1/integrating for l in sj]
                    for k in array_dic.keys(): 
                        if array_dic[k].shape[0] == len(sj):
                            array_dic[k] = np.concatenate([
                                array_dic[k], np.nansum(array_dic[k][levels_to_integrate], axis=0).reshape(1,-1)], axis=0)
                    return array_dic
                φs = __integrate_decomposedarrays_in_dictionary__(φs)
                logger.debug(f'\t\tIntegration took {round(time.time() - info_t_startconditionalsampling)} s (run_wt).')
                
                # Guarantee arrays are 2D inside dictionary (needed for concatenate)
                def __guarantee_2d_arrays_in_dictionary__(array_dic):
                    for k in array_dic.keys(): 
                        if len(array_dic[k].shape) == 1:
                            array_dic[k] = array_dic[k].reshape(1, -1)
                    return array_dic
                φs = __guarantee_2d_arrays_in_dictionary__(φs)
                μs = __guarantee_2d_arrays_in_dictionary__(μs)
                ζb = __guarantee_2d_arrays_in_dictionary__(ζb)
                logger.debug(f'\t\tGuarantee took {round(time.time() - info_t_startconditionalsampling)} s (run_wt).')
                
                φs_names = []
                for n in φs.keys():  
                    if φs[n].shape[0] > 1:
                        for l in sj + ['']: # use if __integrate_decomposedarrays_in_dictionary__
                            if l: φs_names += [f'{n}_{l}'] 
                            else: φs_names += [n] 
                    else: φs_names += [n]            
                # [f'{n}_{l}' if l else n for n in φs.keys() for l in sj + [''] if φs[n].shape[0] > 1] +
                # [n for n in φs.keys() if φs[n].shape[0] <= 1] + 
                logger.debug(f'\t\tTransform 2D arrays to DataFrame with columns `{"`; `".join(φs_names + [f"{n}_qc" for n in μs.keys()] + [f"{n}_nfb" for n in ζb.keys()])}`.')
                logger.debug(f'\t\t{[np.array(v).shape for v in list(φs.values()) + list(μs.values()) + list(ζb.values())]}.')
                logger.debug(f'\t\t{timestamp.shape}.')
                __temp__ = hc24.matrixtotimetable(np.array(timestamp), #data.TIMESTAMP),
                                            np.concatenate(list(φs.values()) + list(μs.values()) + list(ζb.values()), axis=0),
                                            columns=φs_names + [f"{n}_qc" for n in μs.keys()] + [f"{n}_nfb" for n in ζb.keys()])
                logger.debug(f'\t\tMatrix to time table took {round(time.time() - info_t_startconditionalsampling)} s (run_wt).')
                
                ## END HERE FUNCTION RUN_COV 

                #__temp__ = __temp__[__temp__.TIMESTAMP >= min(yl)]
                #__temp__ = __temp__[__temp__.TIMESTAMP < max(yl)]
                
                #__temp__ = __temp__.melt('TIMESTAMP')
                #__temp__ = pd.concat([__temp__.pop('variable').str.extract("^(?P<variable>.+?)_?(?P<natural_frequency>(?<=_)\d+)?$", expand=True), __temp__], axis=1)
                #__temp__['natural_frequency'] = __temp__.variable.apply(lambda x: float(x.rsplit('_', 1)[-1]) if '_' in x and x[-1] in ['0','1','2','3','4','5','6','7','8','9'] else np.nan)
                #__temp__['natural_frequency'] = __temp__['natural_frequency'].apply(lambda j: 1/j2sj(j, 1/dt) if j else np.nan)
                #__temp__['variable'] = __temp__.variable.apply(lambda x: x.rsplit('_', 1)[0] if '_' in x and x[-1] in ['0','1','2','3','4','5','6','7','8','9'] else x)

                """
                def __arr2dataframe__(Y, qc=np.nan, prefix=''.join(xy), 
                                    id=np.array(data.TIMESTAMP), icolnames=sj):
                    if (len(Y.shape) == 1) or (Y.shape[1] == 1): colnames = [f"{prefix}"]
                    else: colnames = ["{}_{}".format(prefix, l) for l in icolnames] if icolnames is not None else None
                    __temp__ = hc24.matrixtotimetable(id, Y, columns=colnames)
                    if qc is not None: __temp__["{}_qc".format(prefix)] = qc
                    __temp__ = __temp__[__temp__.TIMESTAMP > min(yl)]
                    __temp__ = __temp__[__temp__.TIMESTAMP <= max(yl)]
                    return __temp__
                
                __temp__ = reduce(lambda left, right: pd.merge(left, right[['TIMESTAMP'] + list(right.columns.difference(left.columns))], 
                                                            on="TIMESTAMP", how="outer"),
                                [__arr2dataframe__(Y, μs[n], prefix=n) for n, Y in φs.items()] + 
                                [__arr2dataframe__(Y, None,  prefix=f'{n}_nfb') for n, Y in ζb.items()])
                #for k, v in ζb.items(): __temp__[f'{k}_nfb'] = v
                """
                
                info_t_startsaveloop = time.time()
                pattern = re.compile(r"^(?P<variable>.+?)_?(?P<natural_frequency>(?<=_)\d+)?$")
                arr_slice = np.unique(__temp__.TIMESTAMP.dt.floor(str(averaging)+'Min'), return_index=True)
                for __datea__, __tempa__ in list(zip(arr_slice[0], np.split(__temp__, arr_slice[1][1:]))):
                    #for __datea__, __tempa__ in __temp__.groupby(__temp__.TIMESTAMP.dt.floor(str(averaging)+'Min')):
                    dst_path = output_path.format(suffix + "_full_cospectra", pd.to_datetime(__datea__).strftime('%Y%m%d%H%M'))
                    if os.path.exists(dst_path): continue
                    use_header = False
                    if not os.path.exists(dst_path+'.part'):
                        use_header = True
                        header  = "wavelet_based_(co)spectra\n"
                        header += f"--------------------------------------------------------------\n"
                        header += f"TIMESTAMP_START = {min(__tempa__.TIMESTAMP)}\n"
                        header += f"TIMESTAMP_END = {max(__tempa__.TIMESTAMP)}\n"
                        header += f"N: {len(__tempa__.TIMESTAMP)}\n"
                        header += f"TIME_BUFFER [min] = {buffer/60}\n"
                        header += f"frequency [Hz]\n"
                        header += f"y-axis_->_wavelet_coefficient_*_\n"
                        header += f"mother_wavelet -> {method}\n"
                        header += f"acquisition_frequency [Hz] = {1/dt}\n"
                        header += f"averaging_interval [Min] = {averaging}\n"
                        hc24.mkdirs(dst_path)
                        with open(dst_path+'.part', 'w+') as part: part.write(header)
                        legitimate_to_write = 1
                        logger.debug(f'\t\tSaving header of DataFrame took {round(time.time() - info_t_startsaveloop)} ({round(time.time() - info_t_startdataframe)}) s (run_wt).')
                    
                    if not legitimate_to_write: continue
                    
                    __tempa__ = __tempa__.melt('TIMESTAMP')
                    __tempa__.drop('TIMESTAMP', axis=1, inplace=True)
                    __tempa__ = __tempa__.groupby(['variable'], dropna=False).agg(np.nanmean).reset_index(drop=False)
                    __tempa__ = pd.concat([__tempa__.pop('variable').str.extract(pattern, expand=True), __tempa__], axis=1)
                    __tempa__['natural_frequency'] = __tempa__['natural_frequency'].apply(lambda j: 1/j2sj(j, 1/dt) if j else np.nan)
                    with open(dst_path+'.part', 'a+', newline='') as part: __tempa__.to_file(part, header=use_header, chunksize=500, index=False)
                    
                    del __tempa__
                logger.debug(f'\t\tSaving DataFrame for all averaging period took {round(time.time() - info_t_startsaveloop)} ({round(time.time() - info_t_startdataframe)}) s (run_wt).')
                '''
                for a in avg_:
                    __tempa__ = copy.deepcopy(__temp__)
                    #__tempa__["TIMESTAMP"] = pd.to_datetime(np.array(__tempa__.TIMESTAMP)).floor(str(a)+'Min')
                    __tempa__["TIMESTAMP"] = __tempa__.TIMESTAMP.dt.floor(str(a)+'Min')
                    __tempa__ = __tempa__.groupby("TIMESTAMP").agg(np.nanmean).reset_index()

                    """
                    #maincols = ["TIMESTAMP", ''.join(xy)]
                    if φc:
                        #maincols += [names[0] + c + names[1] for c in ['++', '+-', '--', '-+']]
                        sign_label = {1: "+", -1: "-", 0: "·"}
                        for co in set(itertools.combinations([1, -1, 0]*len(names), len(names))):
                            #for c in ['++', '+-', '--', '-+']:
                            sign = ''.join([sign_label[c] for c in co])
                            #name = names[0] + sign[:2] + names[1] + ''.join([s + names[2+i]  for i, s in enumerate(sign[2:])])
                            name = ''.join([c for cs in zip(names, sign) for c in cs])
                            __tempa__.insert(1, name, np.sum(__tempa__[[
                                f"{name}_{l}" for l in sj if dt*2**l < integrating]], axis=1))
                            
                    __tempa__.insert(1, ''.join(xy)+'·', np.sum(__tempa__[[
                        "{}·_{}".format(''.join(xy), l) for l in sj if dt*2**l < integrating]], axis=1))
                    __tempa__.insert(1, ''.join(xy), np.sum(__tempa__[[
                        "{}_{}".format(''.join(xy), l) for l in sj if dt*2**l < integrating]], axis=1))
                    """
                    temporary_name = output_path.rsplit('/', 1)[0] + '/temp/' + output_path.rsplit('/', 1)[1].format(
                        suffix, str(date), str(a).zfill(2)).rsplit('.', 1)[0] + f'.{thisvar_i}.temporary'
                    hc24.mkdirs(temporary_name)
                    dat_fullspectra[a] += [temporary_name]
                    __tempa__.to_file(temporary_name, index=False)
                    #dat_fullspectra[a] += [__tempa__]
                    #dat_fluxresult[a] += [__tempa__[maincols]]
                    del __tempa__
                '''
                del __temp__
                
                    
                logger.debug(f'\t\tMoving from array to DataFrame took {round(time.time() - info_t_startdataframe)} s (run_wt).')
                info_t_dataframe += time.time() - info_t_startdataframe
                print('\x1B[1A\r', f"{date} {np.round(100 * thisvar_i / len(varstorun))}%" + " "*10, end='\n')
            
            logger.info(f'\tDecomposing all {info_c_varstorunloop} variables took {round(info_t_varstorunloop)} s ({round(info_t_varstorunloop/info_c_varstorunloop)} s/var) (run_wt).')
            logger.info(f'\tArray to DataFrame took {round(info_t_dataframe)} s ({round(info_t_dataframe/(thisvar_i+1))} s/loop) (run_wt).')
            logger.info(f'\tStarting averaging at {round(time.time() - info_t_start)} s (run_wt).')
            
            #data.TIMESTAMP
            arr_slice = np.unique(timestamp.dt.floor(str(averaging)+'Min'), return_index=True)
            for __datea__ in arr_slice[0]:
                #for __datea__, _ in data.groupby(data.TIMESTAMP.dt.floor(str(averaging)+'Min')):
                dst_path = output_path.format(suffix + "_full_cospectra", pd.to_datetime(__datea__).strftime('%Y%m%d%H%M'))
                if os.path.exists(dst_path+'.part'): os.rename(dst_path+'.part', dst_path)


            """
            for a in avg_:
                dat_fullspectra_file[a] = reduce(lambda left, right: pd.merge(left, right[['TIMESTAMP'] + list(right.columns.difference(left.columns))],
                                                                        on="TIMESTAMP", how="outer", suffixes=('', '_y')),
                                    [pd.read_file(d) for d in dat_fullspectra[a]])
                #dat_fluxresult[a] = reduce(lambda left, right: pd.merge(left, right[['TIMESTAMP'] + list(right.columns.difference(left.columns))],
                #                                                        on="TIMESTAMP", how="outer", suffixes=('', '_y')),
                #                     dat_fluxresult[a])
                
                hc24.mkdirs(output_path.format(suffix, str(date), str(a).zfill(2)))
                dat_fullspectra_file[a].to_csv(output_path.format(
                    suffix + "_full_cospectra", str(date), str(a).zfill(2)), index=False)
                #dat_fluxresult[a].to_csv(output_path.format(
                #    suffix, str(date), str(a).zfill(2)), index=False)
            """
        except Exception as e:
            logger.critical(str(e))
            print(str(e))
        
        #[os.remove(d) for a in avg_ for d in dat_fullspectra[a] if os.path.exists(d)]
        if os.path.exists(curoutpath_inprog): os.remove(curoutpath_inprog)
        prev_print = '\x1B[2A\r' + f' {date} {len(yl)} files {int(100*ymd_i/len(ymd))} % ({time.strftime("%d.%m.%y %H:%M:%S")})' + '\n'
            
        logger.info(f'\tFinish {date} took {round(time.time() - info_t_startdateloop)} s, yielded {len(yl)} files (run_wt). Progress: {len(yl)} {int(100*ymd_i/len(ymd))} %')
    
    logger.info(f'Finish wavelets at {round(time.time() - info_t_start)} s (run_wt).')
    return



# standard modules
import os
import sys
import re
import logging
import datetime
# Project modules
import yaml
from . import _core as hc24
from ._core import partition as ptt
from ._core import wavelet_functions as wavelet_functions


def __possible_combinations__(interesting_combinations, variables_available):
        # Reduce interesting to possible
        possible_combinations = [sum([v not in variables_available for v in re.split('[*|]', t)])==0 for t in interesting_combinations]
        # Limit run to the realm of possible 
        varstorun = [t for t, p in zip(interesting_combinations, possible_combinations) if p]
        return varstorun


def eddypro_wavelet_run(sitename, inputpath, outputpath, datetimerange, acquisition_frequency=20, fileduration=30, 
         processduration='1D', integratioperiod=None, preaverage=None,
         covariance = None, variables_available=['u', 'v', 'w', 'ts', 'co2', 'h2o'], denoise=0, deadband=[], 
         method = 'dwt', wave_mother='db6', **kwargs):
    local_args = locals()

    if outputpath is not None:
        logname = str(os.path.join(outputpath, f"log/current_{datetime.datetime.now().strftime('%y%m%dT%H%M%S')}.log"))
        hc24.mkdirs(logname)
        #with open(logname, "w+"): pass
        logging.basicConfig(filename=logname,
                            filemode='a',
                            format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S',
                            level=logging.DEBUG, 
                            force=True)

        logging.captureWarnings(True)
        logging.info("STARTING THE RUN")

        # Select output file path
        if method == 'cov':
            outputpath = str(os.path.join(outputpath, str(sitename)+'{}_{}.csv'))
        else:
            outputpath = str(os.path.join(outputpath, 'wavelet_full_cospectra', str(sitename)+'_CDWT{}_{}.csv'))

        # Save args for run
        hc24.mkdirs(outputpath)
        with open(os.path.join(os.path.dirname(os.path.dirname(outputpath)), f'log/setup_{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}.yml'), 'w+') as stp:
            yaml.safe_dump(local_args, stp)

    # Select covariances
    # x*y → Cov(x, y)
    # x*y|x*z|x*... → Cov(x, y)|Cov(x, z),Cov(x, ...)
    if covariance is None:
        interesting_combinations = ['co2*co2', 'h2o*h2o', 'ts*ts', 'co*co',  'ch4*ch4', 'n2o*n2o',
                                    'w*co2', 'w*h2o', 'w*ts', 'w*co', 'w*ch4',  'w*n2o',
                                    'w*co2|w*h2o', 'w*co2|w*co', 'w*co2|w*ch4', 'w*co2|w*ts', 'w*co2|w*h2o|w*co', 
                                    'w*h2o|w*co2', 'w*h2o|w*co', 'w*h2o|w*ch4', 'w*h2o|w*ts', 
                                    'w*co|w*co2',  'w*co|w*ts', 'w*co|w*ch4', 'w*co|w*h2o', 
                                    'w*ch4|w*co2',  'w*ch4|w*co', 'w*ch4|w*ts', 'w*ch4|w*h2o', 
                                    'w*ts|w*co2',  'w*ts|w*co', 'w*ts|w*ch4', 'w*ts|w*h2o',
                                    ]        
        # Reduce interesting to possible
        # Limit run to the realm of possible 
        covariance = __possible_combinations__(interesting_combinations, variables_available)

    # RUN WAVELET FLUX PROCESSING
    # ymd = [START_DATE, END_DATE, FILE_FREQUENCY]
    data = wavelet_functions.load_data_and_loop(ymd = [datetimerange.split('-')[0], datetimerange.split('-')[1], f'{fileduration}min'],
                                         output_path = outputpath,
                                         varstorun = covariance,
                                         averaging = [fileduration],
                                         processing_time_duration = processduration,
                                         method = method,
                                         wt_kwargs = {'fs': acquisition_frequency, 'wavelet': wave_mother},
                                         raw_kwargs = {'path': inputpath,
                                                       'fkwargs': {'dt': 1/acquisition_frequency}},
                                         verbosity=5)
    return data


def integrate_full_spectra_into_file(sitename, outputpath, integratioperiod=60*30, **kwargs):
    # CONCAT INTO SINGLE FILE
    dst_path = os.path.join(outputpath, str(sitename)+f'_CDWT_full_cospectra.csv')
    
    wavelet_functions.integrate_cospectra(os.path.join(outputpath, 'wavelet_full_cospectra'),
                                          1/integratioperiod, '_CDWT_full_cospectra_([0-9]{12}).csv$', dst_path)
    #hc24.concat_into_single_file(
    #    os.path.join(outputpath, 'wavelet_full_cospectra'), str(sitename)+f'_CDWT_full_cospectra.+.{fileduration}mn.csv', 
    #    output_path=dst_path, skiprows=10)
    
def condition_sampling_partition(sitename, outputpath, variables_available=['u', 'v', 'w', 'ts', 'co2', 'h2o'], **kwargs):
    # RUN PARTITIONING
    dst_path = os.path.join(outputpath, str(sitename)+f'_CDWT_full_cospectra.csv')

    h2o_dw_required_variables = ['w','co2','h2o']
    is_lacking_variable = sum([v not in variables_available for v in h2o_dw_required_variables])
    if not is_lacking_variable:
        try:
            ptt.partition_DWCS_H2O(str(dst_path), 
                                        NEE='NEE', GPP='GPP', Reco='Reco', CO2='wco2', 
                                        CO2neg_H2Opos='wco2-wh2o+', 
                                        CO2neg_H2Oneg='wco2-wh2o-', NIGHT=None)\
                                    .filter(['TIMESTAMP', 'NEE', 'GPP', 'Reco'])\
                                    .to_file(os.path.join(outputpath, str(sitename)+f'_CDWT_partitioning_H2O.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
    
    h2o_co_dw_required_variables = ['w','co2','h2o','co']
    is_lacking_variable = sum([v not in variables_available for v in h2o_co_dw_required_variables])
    if not is_lacking_variable:
        try:
            ptt.partition_DWCS_CO(str(dst_path), 
                                        NEE='NEE', GPP='GPP', Reco='Reco', ffCO2='ffCO2',
                                        CO2='wco2', 
                                        CO2neg_H2Opos='wco2-wh2o+', 
                                        CO2neg_H2Oneg='wco2-wh2o-', 
                                        CO2pos_COpos='wco2+wco+', 
                                        CO2pos_COneg='wco2+wco-',
                                        NIGHT=None)\
                                        .filter(['TIMESTAMP', 'NEE', 'GPP', 'Reco', 'ffCO2'])\
                                        .to_file(os.path.join(outputpath, str(sitename)+f'_CDWT_partitioning_H2O_CO.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
    
    co_dw_required_variables = ['w','co2','co']
    is_lacking_variable = sum([v not in variables_available for v in co_dw_required_variables])
    if not is_lacking_variable:
        try:
            ptt.partition_DWCS_CO(str(dst_path), 
                                        NEE='NEE', GPP='GPP', Reco='Reco', ffCO2='ffCO2',
                                        CO2='wco2', 
                                        CO2neg_H2Opos=['wco2-wco+', 'wco2-wco-'], 
                                        CO2neg_H2Oneg=None, 
                                        CO2pos_COpos='wco2+wco+', 
                                        CO2pos_COneg='wco2+wco-',
                                        NIGHT=None)\
                                        .filter(['TIMESTAMP', 'NEE', 'GPP', 'Reco', 'ffCO2'])\
                                        .to_file(os.path.join(outputpath, str(sitename)+f'_CDWT_partitioning_CO.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
        
    ch4_dw_required_variables = ['w','co2','ch4']
    is_lacking_variable = sum([v not in variables_available for v in ch4_dw_required_variables])
    if not is_lacking_variable:
        try:
            ptt.partition_DWCS_CO(str(dst_path), 
                                        NEE='NEE', GPP='GPP', Reco='Reco', ffCO2='ffCO2',
                                        CO2='wco2', 
                                        CO2neg_H2Opos=['wco2-wch4+', 'wco2-wch4-'], 
                                        CO2neg_H2Oneg=None, 
                                        CO2pos_COpos='wco2+wch4+', 
                                        CO2pos_COneg='wco2+wch4-',
                                        NIGHT=None)\
                                        .filter(['TIMESTAMP', 'NEE', 'GPP', 'Reco', 'ffCO2'])\
                                        .to_file(os.path.join(outputpath, str(sitename)+f'_CDWT_partitioning_CH4.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))


def handle_eddypro_setup(**args):
    if args['eddypro']:
        eddypro_setup = hc24.eddypro_tools.read_eddypro_metadata_file(args['eddypro'])
        if not 'sitename' in args.keys() or args['sitename'] is None: args['sitename'] = eddypro_setup['Project']['project_title']
        if not 'inputpath' in args.keys() or args['inputpath'] is None: args['inputpath'] = eddypro_setup['Project']['out_path'] + '/eddypro_raw_datasets/level_6'
        if not 'outputpath' in args.keys() or args['outputpath'] is None: args['outputpath'] = eddypro_setup['Project']['out_path'] + '/wavelet_flux'
        if not 'datetimerange' in args.keys() or args['datetimerange'] is None: args['datetimerange'] = eddypro_setup['Project']['pr_start_date'].replace('-', '') + eddypro_setup['Project']['pr_start_time'].replace(':', '') + '-' + \
            eddypro_setup['Project']['pr_end_date'].replace('-', '') + eddypro_setup['Project']['pr_end_time'].replace(':', '')
        if not 'fileduration' in args.keys() or args['fileduration'] is None: args['fileduration'] = int(eddypro_setup['RawProcess_Settings']['avrg_len'])
    
        if not 'metadata' in args.keys() or args['metadata'] is None: 
            if eddypro_setup['Project']['proj_file']: args['metadata'] = eddypro_setup['Project']['proj_file']
            else: args['metadata'] = args['eddypro'].rsplit('.', 1)[0] + '.metadata'
    
    if args['metadata']:
        eddypro_metad = hc24.eddypro_tools.read_eddypro_metadata_file(args['metadata'])
        if not 'acquisition_frequency' in args.keys() or args['acquisition_frequency'] is None: args['acquisition_frequency'] = int(float(eddypro_metad['Timing']['acquisition_frequency']))
        #if args['fileduration'] is None: args['fileduration'] = int(eddypro_metad['Timing']['file_duration'])
    
    if not 'variables_available' in args.keys() or args['variables_available'] is None:
        if args['eddypro']:
            args['variables_available'] = ['u', 'v', 'w'] + [k for k in ['co2', 'h2o', 'ch4'] if float(eddypro_setup['Project'][f'col_{k}']) > 0]
        if args['metadata']:
            if float(eddypro_setup['Project']['col_n2o']) > 0:
                gas4 =  eddypro_metad['FileDescription'][f"col_{eddypro_setup['Project']['col_n2o']}_variable"]
                if gas4: args['variables_available'] = args['variables_available'] + [gas4]
    return args



# built-in modules
import re
import warnings

# 3rd party modules
import numpy as np
import pandas as pd

##########################################
###     WAVELET-RELATED                            
##########################################

def __wavemother_str_pycwt__(name):
    wavelets = {w.lower(): vars(pycwt.mothers)[w] for w in ['Morlet', 'Paul', 'DOG', 'MexicanHat']}
    mother = wavelets[re.subn('[0-9]', '',  name.lower())[0]]
    if re.findall('[0-9]+', name.lower()): 
        mother = mother(int(re.findall('[0-9]+', name.lower())[0]))
    else:
        mother = mother()
    return mother

try:
    import pywt
    def bufferforfrequency_dwt(N=0, n_=None, fs=20, level=None, f0=None, max_iteration=10**4, wavelet='db6'):
        if level is None and f0 is None: f0 = 1/(2*60*60)  # 18
        lvl = level if level is not None else int(np.ceil(np.log2(fs/f0)))
        if n_ is None: n_ = fs * 60 * 30
        n0 = N
        cur_iteration = 0
        while True:
            n0 += pd.to_timedelta(n_)/pd.to_timedelta("1s") * fs if isinstance(n_, str) else n_
            if lvl <= pywt.dwt_max_level(n0, wavelet):
                break
            cur_iteration += 1
            if cur_iteration > max_iteration:
                warnings.warn('Limit of iterations attained before buffer found. Current buffer allows up to {} levels.'.format(
                    pywt.dwt_max_level(n0, wavelet)))
                break
        return (n0-N) * fs**-1
except Exception as e:
    print(e)

try:
    import pycwt
    def bufferforfrequency(f0, dt=0.05, param=6, mother="MORLET", wavelet=pycwt.Morlet(6)):
        #check if f0 in right units
        # f0 ↴
        #    /\
        #   /  \
        #  /____\
        # 2 x buffer
        
        if isinstance(wavelet, str): wavelet = __wavemother_str_pycwt__(wavelet)
        c = wavelet.flambda() * wavelet.coi()
        n0 = 1 + (2 * (1/f0) * (c * dt)**-1)
        N = int(np.ceil(n0 * dt))
        return N
except Exception as e:
    print(e)



from . import commons
from . import corrections
from . import eddypro_tools
from . import meteo
from . import partition
from . import wavelet_functions
from . import wavelet_misc
from .commons import *
from .meteo import *
